from collections import Counter
from typing import ClassVar, Self
from uuid import UUID

from pydantic import Field, field_validator, model_validator

import gen_epix.casedb.domain.model as model
from gen_epix.casedb.domain import enum, exc
from gen_epix.commondb.domain.command import (
    Command,
    CrudCommand,
    UpdateAssociationCommand,
)
from gen_epix.commondb.util import copy_model_field
from gen_epix.filter.datetime_range import TypedDatetimeRangeFilter

# Non-CRUD


class CaseTypeSetCaseTypeUpdateAssociationCommand(UpdateAssociationCommand):
    ASSOCIATION_CLASS: ClassVar = model.CaseTypeSetMember
    LINK_FIELD_NAME1: ClassVar = "case_type_set_id"
    LINK_FIELD_NAME2: ClassVar = "case_type_id"

    obj_id1: UUID | None = None
    obj_id2: UUID | None = None
    association_objs: list[model.CaseTypeSetMember]


class CaseTypeColSetCaseTypeColUpdateAssociationCommand(UpdateAssociationCommand):
    ASSOCIATION_CLASS: ClassVar = model.CaseTypeColSetMember
    LINK_FIELD_NAME1: ClassVar = "case_type_col_set_id"
    LINK_FIELD_NAME2: ClassVar = "case_type_col_id"

    obj_id1: UUID | None = None
    obj_id2: UUID | None = None
    association_objs: list[model.CaseTypeColSetMember]


class CreateCaseSetCommand(Command):
    """
    Create a new case set and associate it with the specified data collections and
    cases.
    """

    case_set: model.CaseSet = Field(description="The case set to create.")
    data_collection_ids: set[UUID] = Field(
        description="The data collections to associate with the case set, other than the created_in_data_collection. The latter will be removed from the set if present.",
    )
    case_ids: set[UUID] | None = Field(
        description="The cases to associate with the case set upon creation, if any. These cases must have the same case type as the case set.",
        default=None,
    )

    @model_validator(mode="after")
    def _validate_state(self) -> Self:
        self.data_collection_ids.discard(self.case_set.created_in_data_collection_id)
        return self


class ValidateCasesCommand(Command):
    """
    Validate case data and return a validation report.
    """

    case_type_id: UUID = Field(description="The case type ID that the cases belong to.")
    created_in_data_collection_id: UUID = copy_model_field(
        model.CaseValidationReport, "created_in_data_collection_id"
    )
    data_collection_ids: set[UUID] = copy_model_field(
        model.CaseValidationReport, "data_collection_ids"
    )
    is_update: bool = Field(description="Whether this is an update operation.")
    cases: list[model.CaseForCreateUpdate] = Field(description="The cases to validate.")

    @model_validator(mode="after")
    def _validate_cases(self) -> Self:
        if self.created_in_data_collection_id in self.data_collection_ids:
            raise ValueError(
                "The created in data collection ID may not be in the additional data collection IDs."
            )
        if self.is_update and any(x.id is None for x in self.cases):
            raise ValueError("All cases must have an ID when updating cases")
        if not self.is_update and any(x.case_date is None for x in self.cases):
            raise ValueError("All cases must have a case date when creating cases")
        return self


class CreateCasesCommand(ValidateCasesCommand):
    """
    Create the corresponding cases and return them.
    """

    pass


class RetrieveCaseSetStatsCommand(Command):
    """
    Retrieve statistics for a set of case sets.
    """

    case_set_ids: list[UUID] | None = Field(
        default=None,
        description="The case set ids to retrieve stats for, if not all. UNIQUE",
    )


class RetrieveCaseTypeStatsCommand(Command):
    """
    Retrieve statistics for a set of case types.
    """

    case_type_ids: set[UUID] | None = Field(
        default=None,
        description="The case type ids to retrieve stats for, if not all.",
    )
    datetime_range_filter: TypedDatetimeRangeFilter | None = Field(
        default=None,
        description="The datetime range to filter cases by, if any. The key attribute fo the filter should be left empty.",
    )


class RetrieveCompleteCaseTypeCommand(Command):
    """
    Retrieve a complete case type.
    """

    case_type_id: UUID = Field(description="The ID of the case type to retrieve.")


class RetrieveCasesByQueryCommand(Command):
    """
    Retrieve cases based on a query.
    """

    case_query: model.CaseQuery = Field(description="The query to filter cases by.")


class RetrieveCasesByIdCommand(Command):
    """
    Retrieve cases by their IDs.
    """

    case_ids: list[UUID] = Field(
        description="The case ids to retrieve cases for. UNIQUE"
    )

    @field_validator("case_ids", mode="after")
    def _validate_case_ids(cls, value: list[UUID]) -> list[UUID]:
        if len(set(value)) < len(value):
            raise ValueError("Duplicate case ids")
        return value


class RetrieveCaseRightsCommand(Command):
    """
    Retrieve access rights for a set of cases.
    """

    case_ids: list[UUID] = Field(
        description="The case ids to retrieve access for. UNIQUE"
    )

    @field_validator("case_ids", mode="after")
    def _validate_case_ids(cls, value: list[UUID]) -> list[UUID]:
        if len(set(value)) < len(value):
            raise ValueError("Duplicate case ids")
        return value


class RetrieveCaseSetRightsCommand(Command):
    """
    Retrieve access rights for a set of case sets.
    """

    case_set_ids: list[UUID] = Field(
        description="The case set ids to retrieve access for. UNIQUE"
    )

    @field_validator("case_set_ids", mode="after")
    def _validate_case_set_ids(cls, value: list[UUID]) -> list[UUID]:
        if len(set(value)) < len(value):
            raise ValueError("Duplicate case set ids")
        return value


class RetrievePhylogeneticTreeBySequencesCommand(Command):
    """
    Calculate a phylogenetic tree based on a set of sequence IDs, a tree algorithm, and
    a sequence distance protocol.
    """

    tree_algorithm_code: enum.TreeAlgorithmType = Field(
        description="The algorithm to use for constructing the phylogenetic tree."
    )
    seqdb_seq_distance_protocol_id: UUID = Field(
        description="The ID of the sequence distance protocol to use."
    )
    sequence_ids: list[UUID] = Field(
        description="The IDs of the sequences to calculate the phylogenetic tree for."
    )


class RetrievePhylogeneticTreeByCasesCommand(Command):
    """
    Retrieve a phylogenetic tree based on a set of case IDs, a tree algorithm, and
    a genetic distance case type column.
    """

    tree_algorithm: enum.TreeAlgorithmType = Field(
        description="The algorithm to use for constructing the phylogenetic tree."
    )
    genetic_distance_case_type_col_id: UUID = Field(
        description="The ID of the genetic distance case type column to use."
    )
    case_ids: list[UUID] = Field(
        description="The IDs of the cases to calculate the phylogenetic tree for."
    )


class RetrieveGeneticSequenceByCaseCommand(Command):
    """
    Retrieve a set of genetic sequences based on a set of case IDs and a genetic
    sequence case type column.
    """

    genetic_sequence_case_type_col_id: UUID = Field(
        description="The ID of the genetic sequence case type column to use."
    )
    case_ids: list[UUID] = Field(
        description="The IDs of the cases to retrieve genetic sequences for."
    )


class RetrieveGeneticSequenceFastaByCaseCommand(Command):
    """
    Retrieve a set of genetic sequences in FASTA format based on a set of case IDs and a genetic
    sequence case type column. An iterator is returned that yields the FASTA lines.
    """

    genetic_sequence_case_type_col_id: UUID = Field(
        description="The ID of the genetic sequence case type column to use."
    )
    case_ids: list[UUID] = Field(
        description="The IDs of the cases to retrieve genetic sequences for."
    )


class RetrieveAlleleProfileCommand(Command):
    """
    Retrieve a set of allele profiles based on a set of case IDs and a genetic distance
    case type column.
    """

    genetic_distance_case_type_col_id: UUID = Field(
        description="The ID of the genetic distance case type column to use."
    )
    case_ids: list[UUID] = Field(
        description="The IDs of the cases to retrieve allele profiles for."
    )


class CreateReadSetsForCasesCommand(Command):
    """
    Create read sets for a set of cases based on a read set case type column.
    """

    case_read_sets: list[model.CaseReadSet] = Field(
        description="The CaseReadSets describing for which (case_id, case_type_col_id) a ReadSet is to be created. The CaseReadSet case_ids must be unique, the read_set_id must be None and the read_set may not be None.",
    )

    @field_validator("case_read_sets", mode="after")
    def _validate_case_read_sets(
        cls, case_read_sets: list[model.CaseReadSet]
    ) -> list[model.CaseReadSet]:
        case_ids = [x.case_id for x in case_read_sets]
        read_set_ids = [
            x.read_set_id for x in case_read_sets if x.read_set_id is not None
        ]
        if read_set_ids and len(set(read_set_ids)) < len(read_set_ids):
            raise exc.InvalidArgumentsError("CaseReadSets may not contain read_set_id")
        if any(x.read_set is None for x in case_read_sets):
            raise exc.InvalidArgumentsError(
                "CaseReadSets may not have None for read_set "
            )
        duplicate_case_ids = [x for x, y in Counter(case_ids).items() if y > 1]
        if duplicate_case_ids:
            duplicates_str = ", ".join(str(x) for x in duplicate_case_ids)
            raise exc.InvalidArgumentsError(
                f"Some CaseReadSets have identical case_id: {duplicates_str}"
            )
        return case_read_sets


class CreateFileForReadSetCommand(Command):
    """
    Create a file for a read set associated with a case.
    """

    is_fwd: bool = Field(
        description="Whether the file is a forward read file.", default=True
    )
    case_id: UUID = Field(description="The ID of the case the read set belongs to.")
    case_type_col_id: UUID = Field(
        description="The ID of the read set case type column."
    )
    file_content: bytes = Field(description="The content of the file to create.")


class CreateFileForSeqCommand(Command):
    """
    Create a file for a sequence associated with a case.
    """

    case_id: UUID = Field(description="The ID of the case the sequence belongs to.")
    case_type_col_id: UUID = Field(
        description="The ID of the genetic sequence case type column."
    )
    file_content: bytes = Field(description="The content of the file to create.")


class CreateSeqsForCasesCommand(Command):
    """
    Create sequences for a set of cases based on a genetic sequence case type column.
    """

    case_seqs: list[model.CaseSeq] = Field(
        description="The CaseSequences describing for which (case_id, case_type_col_id) a Sequence is to be created."
    )


class RetrieveLibraryPrepProtocolsCommand(Command):
    """
    Retrieve library preparation protocols from seqdb database
    """

    pass


class RetrieveAssemblyProtocolsCommand(Command):
    """
    Retrieve assembly protocols from seqdb database
    """

    pass


# CRUD


class TreeAlgorithmClassCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.TreeAlgorithmClass


class TreeAlgorithmCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.TreeAlgorithm


class GeneticDistanceProtocolCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.GeneticDistanceProtocol


class CaseTypeCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.CaseType


class CaseTypeSetCategoryCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.CaseTypeSetCategory


class CaseTypeSetCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.CaseTypeSet


class CaseTypeSetMemberCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.CaseTypeSetMember


class DimCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.Dim


class ColCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.Col


class CaseTypeColSetCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.CaseTypeColSet


class CaseTypeColSetMemberCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.CaseTypeColSetMember


class CaseTypeColCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.CaseTypeCol


class CaseCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.Case


class CaseDataCollectionLinkCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.CaseDataCollectionLink


class CaseSetCategoryCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.CaseSetCategory


class CaseSetStatusCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.CaseSetStatus


class CaseSetCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.CaseSet


class CaseSetMemberCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.CaseSetMember


class CaseSetDataCollectionLinkCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.CaseSetDataCollectionLink



# pylint: disable=useless-import-alias
from gen_epix import fastapp
from gen_epix.commondb.domain import command as common_command
from gen_epix.commondb.domain import enum as common_enum
from gen_epix.commondb.domain.command import (
    COMMANDS_BY_SERVICE_TYPE as _COMMON_COMMANDS_BY_SERVICE_TYPE,
)
from gen_epix.commondb.domain.command import Command as Command
from gen_epix.commondb.domain.command import ContactCrudCommand as ContactCrudCommand
from gen_epix.commondb.domain.command import CrudCommand as CrudCommand
from gen_epix.commondb.domain.command import (
    DataCollectionCrudCommand as DataCollectionCrudCommand,
)
from gen_epix.commondb.domain.command import (
    DataCollectionSetCrudCommand as DataCollectionSetCrudCommand,
)
from gen_epix.commondb.domain.command import (
    DataCollectionSetDataCollectionUpdateAssociationCommand as DataCollectionSetDataCollectionUpdateAssociationCommand,
)
from gen_epix.commondb.domain.command import (
    DataCollectionSetMemberCrudCommand as DataCollectionSetMemberCrudCommand,
)
from gen_epix.commondb.domain.command import (
    GetIdentityProvidersCommand as GetIdentityProvidersCommand,
)
from gen_epix.commondb.domain.command import (
    IdentifierIssuerCrudCommand as IdentifierIssuerCrudCommand,
)
from gen_epix.commondb.domain.command import InviteUserCommand as InviteUserCommand
from gen_epix.commondb.domain.command import (
    OrganizationAdminPolicyCrudCommand as OrganizationAdminPolicyCrudCommand,
)
from gen_epix.commondb.domain.command import (
    OrganizationCrudCommand as OrganizationCrudCommand,
)
from gen_epix.commondb.domain.command import (
    OrganizationSetCrudCommand as OrganizationSetCrudCommand,
)
from gen_epix.commondb.domain.command import (
    OrganizationSetMemberCrudCommand as OrganizationSetMemberCrudCommand,
)
from gen_epix.commondb.domain.command import (
    OrganizationSetOrganizationUpdateAssociationCommand as OrganizationSetOrganizationUpdateAssociationCommand,
)
from gen_epix.commondb.domain.command import OutageCrudCommand as OutageCrudCommand
from gen_epix.commondb.domain.command import (
    RegisterInvitedUserCommand as RegisterInvitedUserCommand,
)
from gen_epix.commondb.domain.command import (
    RetrieveOrganizationAdminNameEmailsCommand as RetrieveOrganizationAdminNameEmailsCommand,
)
from gen_epix.commondb.domain.command import (
    RetrieveOrganizationContactCommand as RetrieveOrganizationContactCommand,
)
from gen_epix.commondb.domain.command import (
    RetrieveOutagesCommand as RetrieveOutagesCommand,
)
from gen_epix.commondb.domain.command import (
    RetrieveOwnPermissionsCommand as RetrieveOwnPermissionsCommand,
)
from gen_epix.commondb.domain.command import SiteCrudCommand as SiteCrudCommand
from gen_epix.commondb.domain.command import (
    UpdateAssociationCommand as UpdateAssociationCommand,
)
from gen_epix.commondb.domain.command import UpdateUserCommand as UpdateUserCommand
from gen_epix.commondb.domain.command import (
    UpdateUserOwnOrganizationCommand as UpdateUserOwnOrganizationCommand,
)
from gen_epix.commondb.domain.command import UserCrudCommand as UserCrudCommand
from gen_epix.commondb.domain.command import (
    UserInvitationCrudCommand as UserInvitationCrudCommand,
)
from gen_epix.commondb.domain.command.abac import (
    RetrieveOrganizationsUnderAdminCommand as RetrieveOrganizationsUnderAdminCommand,
)
from gen_epix.commondb.domain.command.organization import (
    RetrieveInviteUserConstraintsCommand as RetrieveInviteUserConstraintsCommand,
)
from gen_epix.commondb.domain.command.rbac import (
    RetrieveSubRolesCommand as RetrieveSubRolesCommand,
)
from gen_epix.omopdb.domain import enum
from gen_epix.omopdb.domain.command.omop import (
    CareSiteCrudCommand as CareSiteCrudCommand,
)
from gen_epix.omopdb.domain.command.omop import (
    CdmSourceCrudCommand as CdmSourceCrudCommand,
)
from gen_epix.omopdb.domain.command.omop import CohortCrudCommand as CohortCrudCommand
from gen_epix.omopdb.domain.command.omop import (
    CohortDefinitionCrudCommand as CohortDefinitionCrudCommand,
)
from gen_epix.omopdb.domain.command.omop import (
    ConceptAncestorCrudCommand as ConceptAncestorCrudCommand,
)
from gen_epix.omopdb.domain.command.omop import (
    ConceptClassCrudCommand as ConceptClassCrudCommand,
)
from gen_epix.omopdb.domain.command.omop import ConceptCrudCommand as ConceptCrudCommand
from gen_epix.omopdb.domain.command.omop import (
    ConceptRelationshipCrudCommand as ConceptRelationshipCrudCommand,
)
from gen_epix.omopdb.domain.command.omop import (
    ConceptSynonymCrudCommand as ConceptSynonymCrudCommand,
)
from gen_epix.omopdb.domain.command.omop import (
    ConditionEraCrudCommand as ConditionEraCrudCommand,
)
from gen_epix.omopdb.domain.command.omop import (
    ConditionOccurrenceCrudCommand as ConditionOccurrenceCrudCommand,
)
from gen_epix.omopdb.domain.command.omop import CostCrudCommand as CostCrudCommand
from gen_epix.omopdb.domain.command.omop import (
    DeviceExposureCrudCommand as DeviceExposureCrudCommand,
)
from gen_epix.omopdb.domain.command.omop import DomainCrudCommand as DomainCrudCommand
from gen_epix.omopdb.domain.command.omop import DoseEraCrudCommand as DoseEraCrudCommand
from gen_epix.omopdb.domain.command.omop import DrugEraCrudCommand as DrugEraCrudCommand
from gen_epix.omopdb.domain.command.omop import (
    DrugExposureCrudCommand as DrugExposureCrudCommand,
)
from gen_epix.omopdb.domain.command.omop import (
    DrugStrengthCrudCommand as DrugStrengthCrudCommand,
)
from gen_epix.omopdb.domain.command.omop import (
    FactRelationshipCrudCommand as FactRelationshipCrudCommand,
)
from gen_epix.omopdb.domain.command.omop import (
    LocationCrudCommand as LocationCrudCommand,
)
from gen_epix.omopdb.domain.command.omop import (
    LocationHistoryCrudCommand as LocationHistoryCrudCommand,
)
from gen_epix.omopdb.domain.command.omop import (
    MeasurementCrudCommand as MeasurementCrudCommand,
)
from gen_epix.omopdb.domain.command.omop import (
    MeasurementRelationCrudCommand as MeasurementRelationCrudCommand,
)
from gen_epix.omopdb.domain.command.omop import (
    MetadataCrudCommand as MetadataCrudCommand,
)
from gen_epix.omopdb.domain.command.omop import NoteCrudCommand as NoteCrudCommand
from gen_epix.omopdb.domain.command.omop import NoteNlpCrudCommand as NoteNlpCrudCommand
from gen_epix.omopdb.domain.command.omop import (
    ObservationCrudCommand as ObservationCrudCommand,
)
from gen_epix.omopdb.domain.command.omop import (
    ObservationPeriodCrudCommand as ObservationPeriodCrudCommand,
)
from gen_epix.omopdb.domain.command.omop import (
    PayerPlanPeriodCrudCommand as PayerPlanPeriodCrudCommand,
)
from gen_epix.omopdb.domain.command.omop import PersonCrudCommand as PersonCrudCommand
from gen_epix.omopdb.domain.command.omop import (
    ProcedureOccurrenceCrudCommand as ProcedureOccurrenceCrudCommand,
)
from gen_epix.omopdb.domain.command.omop import (
    ProviderCrudCommand as ProviderCrudCommand,
)
from gen_epix.omopdb.domain.command.omop import (
    RelationshipCrudCommand as RelationshipCrudCommand,
)
from gen_epix.omopdb.domain.command.omop import (
    SourceToConceptMapCrudCommand as SourceToConceptMapCrudCommand,
)
from gen_epix.omopdb.domain.command.omop import (
    SpecimenCrudCommand as SpecimenCrudCommand,
)
from gen_epix.omopdb.domain.command.omop import (
    SurveyConductCrudCommand as SurveyConductCrudCommand,
)
from gen_epix.omopdb.domain.command.omop import (
    VisitDetailCrudCommand as VisitDetailCrudCommand,
)
from gen_epix.omopdb.domain.command.omop import (
    VisitOccurrenceCrudCommand as VisitOccurrenceCrudCommand,
)
from gen_epix.omopdb.domain.command.omop import VocabularyCrudCommand
from gen_epix.omopdb.domain.command.omop import (
    VocabularyCrudCommand as VocabularyCrudCommand,
)

COMMANDS_BY_SERVICE_TYPE: dict[enum.ServiceType, set[type[fastapp.Command]]] = {
    # Specific commands
    enum.ServiceType.ABAC: set(
        _COMMON_COMMANDS_BY_SERVICE_TYPE[common_enum.ServiceType.ABAC]
    )
    | set(),
    enum.ServiceType.OMOP: {
        CareSiteCrudCommand,
        CdmSourceCrudCommand,
        CohortCrudCommand,
        CohortDefinitionCrudCommand,
        ConceptAncestorCrudCommand,
        ConceptClassCrudCommand,
        ConceptCrudCommand,
        ConceptRelationshipCrudCommand,
        ConceptSynonymCrudCommand,
        ConditionEraCrudCommand,
        ConditionOccurrenceCrudCommand,
        CostCrudCommand,
        DeviceExposureCrudCommand,
        DomainCrudCommand,
        DoseEraCrudCommand,
        DrugEraCrudCommand,
        DrugExposureCrudCommand,
        DrugStrengthCrudCommand,
        FactRelationshipCrudCommand,
        LocationCrudCommand,
        LocationHistoryCrudCommand,
        MeasurementCrudCommand,
        MeasurementRelationCrudCommand,
        MetadataCrudCommand,
        NoteCrudCommand,
        NoteNlpCrudCommand,
        ObservationCrudCommand,
        ObservationPeriodCrudCommand,
        PayerPlanPeriodCrudCommand,
        PersonCrudCommand,
        ProcedureOccurrenceCrudCommand,
        ProviderCrudCommand,
        RelationshipCrudCommand,
        SourceToConceptMapCrudCommand,
        SpecimenCrudCommand,
        SurveyConductCrudCommand,
        VisitDetailCrudCommand,
        VisitOccurrenceCrudCommand,
        VocabularyCrudCommand,
    },
    # Common commands
    enum.ServiceType.AUTH: set(
        _COMMON_COMMANDS_BY_SERVICE_TYPE[common_enum.ServiceType.AUTH]
    ),
    enum.ServiceType.SYSTEM: set(
        _COMMON_COMMANDS_BY_SERVICE_TYPE[common_enum.ServiceType.SYSTEM]
    ),
    enum.ServiceType.RBAC: set(
        _COMMON_COMMANDS_BY_SERVICE_TYPE[common_enum.ServiceType.RBAC]
    ),
    enum.ServiceType.ORGANIZATION: set(
        _COMMON_COMMANDS_BY_SERVICE_TYPE[common_enum.ServiceType.ORGANIZATION]
    ),
}

COMMON_COMMAND_MAP: dict[type[fastapp.Command], type[fastapp.Command]] = {
    common_command.UserCrudCommand: UserCrudCommand,
    common_command.UserInvitationCrudCommand: UserInvitationCrudCommand,
    common_command.InviteUserCommand: InviteUserCommand,
    common_command.UpdateUserCommand: UpdateUserCommand,
    common_command.OrganizationAdminPolicyCrudCommand: OrganizationAdminPolicyCrudCommand,
}



from collections.abc import Hashable, Iterable

from gen_epix.commondb.repositories import (
    OrganizationDictRepository as CommonOrganizationDictRepository,
)
from gen_epix.fastapp import Entity
from gen_epix.seqdb.domain import model
from gen_epix.seqdb.domain.model import Model


class OrganizationDictRepository(CommonOrganizationDictRepository):
    def __init__(
        self,
        entities: Iterable[Entity],
        db: dict[type[Model], dict[Hashable, Model]],
        **kwargs: Any,
    ):
        super().__init__(
            entities,
            db,
            **kwargs,
            user_class=model.User,
            user_invitation_class=model.UserInvitation,
        )



"""
Examples demonstrating usage of the transformer framework.
"""

from pydantic import BaseModel

from gen_epix.transform import (
    ConditionalTransformer,
    FieldTransformer,
    ObjectAdapter,
    Pipeline,
    StreamingPipeline,
    ValidationTransformer,
    register_transformer,
)
from gen_epix.transform.transformer import Transformer


class Person(BaseModel):
    """Example Pydantic model."""

    name: str
    age: int
    email: str


@register_transformer("string_upper")
class StringUpperTransformer(Transformer):
    """Example custom transformer."""

    def __init__(self, name: str = "StringUpperTransformer"):
        super().__init__(name)

    def transform(self, obj: ObjectAdapter) -> ObjectAdapter:
        # This would need to implement the actual transformation
        return obj


def example_usage() -> None:
    """Demonstrate basic usage of the transformer framework."""

    # Sample data - mix of dict and Pydantic models
    data = [
        {"name": "john doe", "age": 30, "email": "JOHN@EXAMPLE.COM"},
        {"name": "jane smith", "age": 25, "email": "jane@EXAMPLE.com"},
        Person(name="bob wilson", age=35, email="BOB@example.COM"),
    ]

    # Create transformers
    name_normalizer = FieldTransformer(
        field_name="name", transform_fn=lambda x: str(x).title(), name="NameNormalizer"
    )

    email_normalizer = FieldTransformer(
        field_name="email",
        transform_fn=lambda x: str(x).lower(),
        name="EmailNormalizer",
    )

    age_validator = ValidationTransformer(
        validator=lambda obj: obj.get("age", 0) > 0, name="AgeValidator"
    )

    # Create pipeline
    pipeline = Pipeline([age_validator, name_normalizer, email_normalizer])

    # Process data
    streaming_pipeline = StreamingPipeline(pipeline)
    successes, errors = streaming_pipeline.collect_errors(iter(data))

    print(f"Successfully processed {len(successes)} objects")
    print(f"Failed to process {len(errors)} objects")

    for success in successes:
        print(f"Success: {success}")

    for error in errors:
        print(f"Error: {error}")


def example_conditional_transformation() -> None:
    """Demonstrate conditional transformation."""

    data = [
        {"name": "John", "country": "US", "phone": "1234567890"},
        {"name": "Alice", "country": "UK", "phone": "9876543210"},
    ]

    # Phone normalization only for US numbers
    us_phone_normalizer = ConditionalTransformer(
        condition=lambda obj: obj.get("country") == "US",
        transformer=FieldTransformer(
            field_name="phone", transform_fn=lambda x: f"+1-{x[:3]}-{x[3:6]}-{x[6:]}"
        ),
        name="USPhoneNormalizer",
    )

    pipeline = Pipeline([us_phone_normalizer])

    for result in pipeline.process_stream(iter(data)):
        if result.success:
            print(f"Transformed: {result.transformed_object}")
        else:
            print(f"Failed: {result.error}")


if __name__ == "__main__":
    print("=== Basic Usage ===")
    example_usage()

    print("\n=== Conditional Transformation ===")
    example_conditional_transformation()



# pylint: disable=useless-import-alias
from gen_epix.transform.adapter import DictAdapter as DictAdapter
from gen_epix.transform.adapter import ObjectAdapter as ObjectAdapter
from gen_epix.transform.adapter import PolarsAdapter as PolarsAdapter
from gen_epix.transform.adapter import PydanticAdapter as PydanticAdapter
from gen_epix.transform.enum import TransformResultType as TransformResultType
from gen_epix.transform.pipeline import FallbackTransformer as FallbackTransformer
from gen_epix.transform.pipeline import Pipeline as Pipeline
from gen_epix.transform.pipeline import RetryTransformer as RetryTransformer
from gen_epix.transform.registry import Registry as Registry
from gen_epix.transform.registry import register_factory as register_factory
from gen_epix.transform.registry import register_transformer as register_transformer
from gen_epix.transform.stream_processer import StreamProcessor as StreamProcessor
from gen_epix.transform.streaming_pipeline import StreamingPipeline as StreamingPipeline
from gen_epix.transform.transform_result import TransformResult as TransformResult
from gen_epix.transform.transformer import Transformer as Transformer
from gen_epix.transform.transformers import (
    ConditionalTransformer as ConditionalTransformer,
)
from gen_epix.transform.transformers import FieldTransformer as FieldTransformer
from gen_epix.transform.transformers import IsoTimeTransformer as IsoTimeTransformer
from gen_epix.transform.transformers import (
    MultiFieldTransformer as MultiFieldTransformer,
)
from gen_epix.transform.transformers import ObjectTransformer as ObjectTransformer
from gen_epix.transform.transformers import TupleMapTransformer as TupleMapTransformer
from gen_epix.transform.transformers import (
    ValidationTransformer as ValidationTransformer,
)



import abc
import uuid
from collections.abc import Callable, Hashable, Iterable
from itertools import chain
from typing import Any

from gen_epix.fastapp import exc
from gen_epix.fastapp.enum import CrudOperation
from gen_epix.fastapp.model import Model
from gen_epix.fastapp.unit_of_work import BaseUnitOfWork
from gen_epix.filter import Filter


class BaseRepository(abc.ABC):
    def __init__(self, **kwargs: Any):
        self._id: str = kwargs.get("id", str(uuid.uuid4()))
        self._name: str = kwargs.get("name", self._id)

    @property
    def id(self) -> str:
        return self._id

    @property
    def name(self) -> str:
        return self._name

    @classmethod
    @abc.abstractmethod
    def create_repository(cls, **kwargs: Any) -> "BaseRepository":
        raise NotImplementedError()

    @abc.abstractmethod
    def crud(
        self,
        uow: BaseUnitOfWork,
        user_id: Hashable | None,
        model_class: type[Model],
        objs: Model | Iterable[Model] | None,
        obj_ids: Hashable | Iterable[Hashable] | None,
        operation: CrudOperation,
        filter: Filter | None = None,
        **kwargs: Any,
    ) -> Hashable | list[Hashable] | Model | list[Model] | bool | list[bool] | None:
        """
        Perform CRUD operations on the repository within a unit of work context. The
        user_id corresponds to the user that executes the operation and can e.g. be
        used to determine permissions of the user or to store the executing user as
        metadata.
        The filter can be used to filter the results of the operation, and is only
        applied to read and delete all operations. The filter must have keys that
        correspond to the row class fields and must be convertable into a where clause,
        and can e.g. be created by the split_filter method.
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def read_fields(
        self,
        uow: BaseUnitOfWork,
        user_id: Hashable | None,
        model_class: type[Model],
        field_names: list[str],
        filter: Filter | None = None,
        **kwargs: Any,
    ) -> Iterable[tuple]:
        """
        Reads specific fields of objects of the given model class that match the
        provided filter, if any. The filter must have keys that correspond to the model class
        fields and must be convertable into a where clause, and can e.g. be created by
        the split_filter method.
        An iterable of tuples is returned, where each tuple contains the values of the
        requested fields for one object.
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def split_filter(
        self, model_class: type, filter: Filter | None
    ) -> tuple[Filter | None, Filter | None]:
        """
        Splits a filter into two parts: a repository filter that can can be applied by
        this repository directly and presumably efficiently (e.g. as a where clause on
        an SQL database) and a service filter with the remainder that cannot and would
        have to be applied by the service after retrieving results filtered by the
        first part. The input filter must have keys that correspond to the model class
        fields.
        The filters are returned as a tuple[repository_filter, service_filter]. In
        case the filter cannot be split, the first element of the tuple will be None.
        In case the entire filter can be converted into a repository_filter, the second
        element of the tuple will be None.
        """
        raise NotImplementedError()

    def update_association(
        self,
        uow: BaseUnitOfWork,
        user_id: Hashable | None,
        model_class: type[Model],
        link_field_name1: str,
        link_field_name2: str,
        obj_id1: Hashable | None,
        obj_id2: Hashable | None,
        association_objs: Iterable[Model],
        **kwargs: Any,
    ) -> list[Model] | list[Hashable]:
        return_id = kwargs.pop("return_id", False)
        excluded_association_objs: Iterable[Model] = kwargs.pop(
            "excluded_association_objs", []
        )

        def get_id_pair(obj: Model) -> tuple[Hashable, Hashable]:
            return (getattr(obj, link_field_name1), getattr(obj, link_field_name2))

        # Parse input and create some general functions and values
        association_objs = list(association_objs)
        assert model_class.ENTITY is not None
        id_field_name = model_class.ENTITY.id_field_name
        if not isinstance(id_field_name, str):
            raise exc.RepositoryServiceError(
                f"Model {model_class.__name__}: id_field_name other than string not implemented"
            )
        get_association_id: Callable[[Model], Hashable] = lambda x: getattr(
            x, id_field_name
        )
        excluded_obj_ids = {get_association_id(x) for x in excluded_association_objs}
        excluded_id_pairs = {get_id_pair(x) for x in excluded_association_objs}

        # Special case: no association objects, i.e. delete all associations of either obj1 or obj2
        if not association_objs:
            # Go over obj_id1 and obj_id2, only one of both should be provided
            for obj_id, link_field_name in zip(
                [obj_id1, obj_id2], [link_field_name1, link_field_name2]
            ):
                if obj_id is None:
                    continue
                to_delete_obj_ids = [
                    get_association_id(x)  # type: ignore
                    for x in self.crud(  # type: ignore
                        uow,
                        user_id,
                        model_class,
                        None,
                        None,
                        CrudOperation.READ_ALL,
                    )
                    if getattr(x, link_field_name) == obj_id
                    and get_association_id(x) not in excluded_obj_ids  # type: ignore
                ]
                self.crud(
                    uow,
                    user_id,
                    model_class,
                    None,
                    to_delete_obj_ids,
                    CrudOperation.DELETE_SOME,
                )
            return []

        # Get obj_id_pairs and verify uniqueness
        obj_id_pairs = [get_id_pair(x) for x in association_objs]
        obj_dict = dict(zip(obj_id_pairs, association_objs))
        if len(obj_id_pairs) != len(frozenset(obj_id_pairs)):
            invalid_ids = list(
                chain.from_iterable(
                    [x for x in obj_id_pairs if obj_id_pairs.count(x) > 1]
                )
            )
            raise exc.DuplicateIdsError(
                f"Model {model_class.__name__}: object id pairs are not unique",
                ids=invalid_ids,
            )

        # Verify if any excluded ids or id pairs are present
        if excluded_obj_ids:
            invalid_association_obj_ids = [
                get_association_id(x)
                for x in association_objs
                if get_association_id(x) in excluded_obj_ids
            ]
            if invalid_association_obj_ids:
                raise exc.InvalidModelIdsError(
                    f"Model {model_class.__name__}: association objects contains some excluded ids",
                    ids=invalid_association_obj_ids,
                )
        if excluded_id_pairs:
            invalid_id_pairs = [x for x in obj_id_pairs if x in excluded_id_pairs]
            if invalid_id_pairs:
                raise exc.InvalidModelIdsError(
                    f"Model {model_class.__name__}: association object id pairs contains some excluded pairs",
                )

        # Retrieve non-excluded existing objects
        # TODO: replace retrieval of existing objects with a more efficient method
        existing_objs: list[Model] = [
            x  # type: ignore
            for x in self.crud(  # type: ignore
                uow, user_id, model_class, None, None, CrudOperation.READ_ALL
            )
            if get_id_pair(x) in obj_dict  # type: ignore
        ]
        existing_obj_dict: dict[Hashable, Model] = {
            get_id_pair(x): x
            for x in existing_objs
            if get_id_pair(x) not in excluded_id_pairs
        }

        # Determine association objects to create, update or delete
        to_create_objs = [y for x, y in obj_dict.items() if x not in existing_obj_dict]
        to_update_objs = [y for x, y in obj_dict.items() if x in existing_obj_dict]
        to_delete_objs = [y for x, y in existing_obj_dict.items() if x not in obj_dict]

        # Verify that objects to create have an id
        if any([get_association_id(x) is None for x in to_create_objs]):
            raise exc.InvalidModelIdsError(
                f"Model {model_class.__name__}: object(s) to create have no id",
                ids=[
                    get_association_id(x)
                    for x in to_create_objs
                    if get_association_id(x) is None
                ],
            )

        # Create association objects
        if to_create_objs:
            self.crud(
                uow,
                user_id,
                model_class,
                to_create_objs,
                None,
                CrudOperation.CREATE_SOME,
            )

        # Update association objects
        to_update_obj_ids = [
            get_association_id(existing_obj_dict[get_id_pair(x)])
            for x in to_update_objs
        ]
        if to_update_obj_ids:
            for obj, obj_id in zip(to_update_objs, to_update_obj_ids):
                # Set actual ids of objects to update, so that this is also corrected in the caller
                setattr(obj, id_field_name, obj_id)
            self.crud(
                uow,
                user_id,
                model_class,
                to_update_objs,
                None,
                CrudOperation.UPDATE_SOME,
            )

        # Delete association objects
        if to_delete_objs:
            self.crud(
                uow,
                user_id,
                model_class,
                None,
                [get_association_id(x) for x in to_delete_objs],
                CrudOperation.DELETE_SOME,
            )

        return (
            [get_association_id(x) for x in association_objs]
            if return_id
            else association_objs
        )

    @abc.abstractmethod
    def verify_valid_ids(
        self,
        uow: BaseUnitOfWork,
        user_id: Hashable,
        model_class: type[Model],
        obj_ids: Iterable[Hashable],
        verify_exists: bool = True,
        verify_duplicate: bool = True,
    ) -> None:
        raise NotImplementedError

    @abc.abstractmethod
    def uow(self, **kwargs: Any) -> BaseUnitOfWork:
        raise NotImplementedError()

    @staticmethod
    def raise_on_duplicate_ids(obj_ids: Iterable[Hashable]) -> None:
        set_ = set()
        duplicate_ids = [x for x in obj_ids if x in set_ or set_.add(x)]  # type: ignore
        if duplicate_ids:
            duplicate_ids_str = ", ".join([f'"{x}"' for x in duplicate_ids])
            raise exc.DuplicateIdsError(
                f"Object ids are not unique: {duplicate_ids_str}",
                ids=duplicate_ids,
            )

    @staticmethod
    def verify_crud_args(
        model_class: type[Model],
        objs: Model | Iterable[Model] | None,
        obj_ids: Hashable | Iterable[Hashable] | None,
        operation: CrudOperation,
    ) -> None:

        def _verify_no_objs() -> None:
            if objs is not None:
                raise exc.InvalidArgumentsError(
                    "Invalid objs: expected None, " f"got {type(objs)}"
                )

        def _verify_no_obj_ids() -> None:
            if obj_ids is not None:
                raise exc.InvalidArgumentsError(
                    "Invalid obj_ids: expected None, " f"got {type(obj_ids)}"
                )

        def _verify_no_data() -> None:
            _verify_no_objs()
            _verify_no_obj_ids()

        def _verify_one_obj() -> None:
            if not isinstance(objs, model_class):
                raise exc.InvalidArgumentsError(
                    "Invalid objs: expected single object, " f"got {type(objs)}"
                )
            _verify_no_obj_ids()

        def _verify_one_id() -> None:
            _verify_no_objs()
            if not isinstance(obj_ids, Hashable):
                raise exc.InvalidArgumentsError(
                    "Invalid obj_ids: expected single obj_id, " f"got {type(obj_ids)}"
                )

        def _verify_some_objs() -> None:
            if not isinstance(objs, Iterable) or isinstance(objs, model_class):
                raise exc.InvalidArgumentsError(
                    "Invalid objs: expected iterable of objs, " f"got {type(objs)}"
                )
            _verify_no_obj_ids()

        def _verify_some_ids() -> None:
            _verify_no_objs()
            if not isinstance(obj_ids, Iterable):
                raise exc.InvalidArgumentsError(
                    "Invalid obj_ids: expected iterable of obj_ids, "
                    f"got {type(obj_ids)}"
                )

        match operation:
            case CrudOperation.CREATE_ONE:
                _verify_one_obj()
            case CrudOperation.CREATE_SOME:
                _verify_some_objs()
            case CrudOperation.READ_ONE:
                _verify_one_id()
            case CrudOperation.READ_SOME:
                _verify_some_ids()
            case CrudOperation.READ_ALL:
                _verify_no_data()
            case CrudOperation.UPDATE_ONE:
                _verify_one_obj()
            case CrudOperation.UPDATE_SOME:
                _verify_some_objs()
            case CrudOperation.UPSERT_ONE:
                _verify_one_obj()
            case CrudOperation.UPSERT_SOME:
                _verify_some_objs()
            case CrudOperation.DELETE_ONE:
                _verify_one_id()
            case CrudOperation.DELETE_SOME:
                _verify_some_ids()
            case CrudOperation.DELETE_ALL:
                _verify_no_data()
            case CrudOperation.UNDELETE_ONE:
                _verify_one_id()
            case CrudOperation.UNDELETE_SOME:
                _verify_some_ids()
            case CrudOperation.UNDELETE_ALL:
                _verify_no_data()
            case CrudOperation.RESTORE_ONE:
                _verify_one_id()
            case CrudOperation.RESTORE_SOME:
                _verify_some_ids()
            case CrudOperation.EXISTS_ONE:
                _verify_one_id()
            case CrudOperation.EXISTS_SOME:
                _verify_some_ids()
            case _:
                raise NotImplementedError(f"Operation {operation} not implemented")



import abc
from collections.abc import Callable, Hashable, Iterable, Iterator
from typing import Annotated, Any, Literal, Self

from pydantic import BaseModel, Field, WithJsonSchema

from gen_epix.filter.enum import FilterType


class Filter(BaseModel):
    """
    Base class for filters.

    Attributes:
        invert (bool): Whether to invert the filter.
        key (Hashable | None): The column key to apply the filter to, when applied to a row. If None, the filter cannot be applied to a row, only to a column.
    """

    invert: bool = Field(default=False, description="Whether to invert the filter.")
    key: Annotated[
        Hashable | None,
        WithJsonSchema({"type": "string"}),
    ] = Field(
        default=None,
        description="The column key to apply the filter to, when applied to a row. If None, the filter cannot be applied to a row, only to a column.",
        json_schema_extra={"type": "string"},
    )
    _is_composite: bool = False

    @property
    def is_composite(self) -> bool:
        return self._is_composite

    def _get_row_value(
        self, row: dict | BaseModel, key: Hashable, is_model: bool
    ) -> Any:
        """Helper method to get a attribute value from a row model"""
        if is_model:
            return getattr(row, key, None)
        return row.get(key, None)

    def match_value(
        self,
        value: Any | None,
        na_values: set[Any] | None = None,
        map_fn: Callable[[Any], Any] | None = None,
    ) -> bool:
        """
        Check if a value matches the filter.

        Args:
            value (Any | None): The value to be checked.
            na_values (set[Any] | None, optional): Set of values to be considered as NA values. Defaults to None.
            map_fn (Callable[[Any], Any] | None, optional): Function to be applied to the value before matching. Defaults to None.

        Returns:
            bool: True if the value matches the filter, False otherwise.
        """
        if not map_fn:
            map_fn = lambda x: x
        if na_values is None:
            return (value is not None and self._match(map_fn(value))) ^ self.invert
        else:
            return (value not in na_values and self._match(map_fn(value))) ^ self.invert

    def match_column(
        self,
        values: Iterable[Any | None],
        na_values: set[Any] | None = None,
        map_fn: Callable[[Any], Any] | None = None,
    ) -> Iterator[bool]:
        """
        Check if each value in a column matches the filter.

        Args:
            values (Iterable[Any | None]): The values in the column.
            na_values (set[Any] | None, optional): Set of values to be considered as NA values. Defaults to None.
            map_fn (Callable[[Any], Any] | None, optional): Function to be applied to each value before matching. Defaults to None.

        Yields:
            bool: True if the value matches the filter, False otherwise.
        """
        if not map_fn:
            map_fn = lambda x: x
        if na_values is None:
            for value in values:
                yield (value is not None and self._match(map_fn(value))) ^ self.invert
        else:
            for value in values:
                yield (value not in na_values) ^ self.invert

    def filter_column(
        self,
        values: Iterable[Any | None],
        na_values: set[Any] | None = None,
        map_fn: Callable[[Any], Any] | None = None,
    ) -> Iterator[Any | None]:
        """
        Analogous to match_column, but yields the values that match the filter instead of a bool.
        """
        if not map_fn:
            map_fn = lambda x: x
        if na_values is None:
            for value in values:
                if (value is not None and self._match(map_fn(value))) ^ self.invert:
                    yield value
        else:
            for value in values:
                if (value not in na_values) ^ self.invert:
                    yield value

    def match_row(
        self,
        row: dict[Hashable, Any | None] | BaseModel,
        na_values: set[Any] | None = None,
        map_fn: Callable[[Any], Any] | None = None,
        is_model: bool = False,
    ) -> bool:
        """
        Check if a row matches the filter.

        Args:
            row (dict[Hashable, Any | None]): The row to be checked.
            na_values (set[Any] | None, optional): Set of values to be considered as NA values. Defaults to None.
            map_fn (Callable[[Any], Any] | None, optional): Function to be applied to each value before matching. Defaults to None.

        Returns:
            bool: True if the row matches the filter, False otherwise.
        """
        # Match if both key exists, value not null and value matches
        if self.key is None:
            raise ValueError("Key must be set to apply filter to a row.")
        if not map_fn:
            map_fn = lambda x: x
        key = self.key
        if na_values is None:
            return (
                (is_model or key in row)
                and self._get_row_value(row, key, is_model) is not None
                and self._match(map_fn(self._get_row_value(row, key, is_model)))
            ) ^ self.invert
        else:
            return (
                (is_model or key in row)
                and self._get_row_value(row, key, is_model) not in na_values
                and self._match(map_fn(self._get_row_value(row, key, is_model)))
            ) ^ self.invert

    def match_rows(
        self,
        rows: Iterable[dict[Hashable, Any | None] | BaseModel],
        na_values: set[Any] | None = None,
        map_fn: Callable[[Any], Any] | None = None,
        is_model: bool = False,
    ) -> Iterator[bool]:
        """
        Check if each row in a collection of rows matches the filter.

        Args:
            rows (Iterable[dict[Hashable, Any | None]]): The collection of rows.
            na_values (set[Any] | None, optional): Set of values to be considered as NA values. Defaults to None.
            map_fn (Callable[[Any], Any] | None, optional): Function to be applied to each value before matching. Defaults to None.

        Yields:
            bool: True if the row matches the filter, False otherwise.
        """
        # Match if both key exists, value not null and value matches
        if self.key is None:
            raise ValueError("Key must be set to apply filter to a row.")
        if not map_fn:
            map_fn = lambda x: x
        key = self.key
        if na_values is None:
            for row in rows:
                yield (
                    (is_model or key in row)
                    and self._get_row_value(row, key, is_model) is not None
                    and self._match(map_fn(self._get_row_value(row, key, is_model)))
                ) ^ self.invert
        else:
            for row in rows:
                yield (
                    (is_model or key in row)
                    and self._get_row_value(row, key, is_model) not in na_values
                    and self._match(map_fn(self._get_row_value(row, key, is_model)))
                ) ^ self.invert

    def filter_rows(
        self,
        rows: Iterable[dict[Hashable, Any | None] | BaseModel],
        na_values: set[Any] | None = None,
        map_fn: Callable[[Any], Any] | None = None,
        is_model: bool = False,
    ) -> Iterator[dict[Hashable, Any | None] | BaseModel]:
        """
        Analogous to match_rows, but yields the rows that match the filter instead of a bool.
        """
        # Match if both key exists, value not null and value matches
        if self.key is None:
            raise ValueError("Key must be set to apply filter to a row.")
        if not map_fn:
            map_fn = lambda x: x
        key = self.key
        if na_values is None:
            for row in rows:
                if (
                    (is_model or key in row)
                    and self._get_row_value(row, key, is_model) is not None
                    and self._match(map_fn(self._get_row_value(row, key, is_model)))
                ) ^ self.invert:
                    yield row
        else:
            for row in rows:
                if (
                    (is_model or key in row)
                    and self._get_row_value(row, key, is_model) not in na_values
                    and self._match(map_fn(self._get_row_value(row, key, is_model)))
                ) ^ self.invert:
                    yield row

    @abc.abstractmethod
    def _match(self, value: Any) -> bool:
        """
        Abstract method to be implemented by subclasses.
        Checks if a value matches the filter.

        Args:
            value (Any): The value to be checked.

        Returns:
            bool: True if the value matches the filter, False otherwise.
        """
        raise NotImplementedError()

    def set_key(self, key: Hashable | Callable[[Hashable], Hashable]) -> Self:
        if callable(key):
            self.key = key(self.key)
        else:
            self.key = key
        return self

    def get_key(self) -> Hashable:
        return self.key

    def __call__(
        self,
        data: Iterable[Any],
        axis: int = 1,
        na_values: set[Any] | None = None,
        map_fn: Callable[[Any], Any] | None = None,
    ) -> Iterator[Any]:
        """
        Apply filter to data.

        Args:
            data (Iterable[Any]): The data to be filtered.
            axis (int, optional): The axis along which to apply the filter. 0 for rows, 1 for columns. Defaults to 1.
            na_values (set[Any] | None, optional): Set of values to be considered as NA values. Defaults to None.
            map_fn (Callable[[Any], Any] | None, optional): Function to be applied to each value before matching. Defaults to None.

        Returns:
            Iterable[bool]: An iterable of booleans for each row, True if the data matches the filter, False otherwise.
        """
        if axis == 0:
            return self.filter_rows(data, na_values=na_values, map_fn=map_fn)
        if axis == 1:
            return self.filter_column(data, na_values=na_values, map_fn=map_fn)
        raise ValueError("Axis must be 0 or 1.")


class TypedFilter(Filter):
    type: Literal[FilterType.BASE.value]



"""Settings manager for handling application settings."""

import json
import os
from collections.abc import Callable
from pathlib import Path
from typing import Any

from dynaconf import Dynaconf


class SettingsManager:
    """Manages application settings with environment variable overrides."""

    DEFAULT_SETTINGS_FILES_ENVVAR = "SETTINGS_FILES"
    DEFAULT_ENVVAR_SEPARATOR = "__"

    def __init__(self, prefix: str, lowercase_keys: bool = True):
        """
        Initialize settings manager.
        """
        self.prefix = prefix
        self.prefix_without_underscore = prefix.rstrip("_")  # necessary for dynaconf
        self._settings_cache: Dynaconf | None = None
        self.lowercase_keys = lowercase_keys

    def load_settings(
        self,
        settings_files: list[str] | str | None = None,
        settings_files_envvar: str | None = None,
        envvar_separator: str | None = None,
        post_hooks: Callable | list[Callable] | None = None,
    ) -> Dynaconf:
        """
        Load settings from one or more settings file(s) specified either as a
        list or if not, in an environment variable.
        """
        settings_files_envvar = (
            settings_files_envvar or self.DEFAULT_SETTINGS_FILES_ENVVAR
        )
        envvar_separator = envvar_separator or self.DEFAULT_ENVVAR_SEPARATOR
        # Determine settings files to use
        if settings_files is None:
            # Try to get from environment variable
            settings_files = os.environ.get(f"{self.prefix}{settings_files_envvar}")
            if settings_files:
                settings_files = SettingsManager.parse_settings_files_from_string(
                    settings_files
                )
        elif isinstance(settings_files, list):
            pass
        else:
            settings_files = SettingsManager.parse_settings_files_from_string(
                settings_files
            )
        if isinstance(settings_files, str):
            settings_files = [settings_files]
        if not settings_files:
            raise ValueError(
                "No settings files provided. Specify via argument or "
                f"environment variable {self.prefix}{settings_files_envvar}"
            )
        for settings_file_path in settings_files:
            if not Path(settings_file_path).exists():
                raise FileNotFoundError(
                    f"Custom settings file not found: {settings_file_path}"
                )

        # Load settings using dynaconf for environment variable support
        settings = Dynaconf(
            envvar_prefix=self.prefix_without_underscore,
            settings_files=settings_files,
            envvar_separator=envvar_separator,  # Support nested keys like API__HOST
            lowercase_read=self.lowercase_keys,  # Ensure keys are lowercase for Pydantic
            ignore_unknown_envvars=True,
            merge_enabled=True,
            post_hooks=post_hooks,
        )
        self._settings_cache = settings

        return self._settings_cache

    @property
    def settings(self) -> Dynaconf:
        """Get loaded settings."""
        if self._settings_cache is None:
            raise RuntimeError("Settings not loaded. Call load_settings() first.")
        return self._settings_cache

    def get_setting(self, key_path: str, default: Any = None) -> Any:
        """Get setting value by dot-notation path.

        Args:
            key_path: Dot-separated path to setting (e.g., 'api.host')
            default: Default value if not found

        Returns:
            Setting value or default
        """
        if self._settings_cache is None:
            raise RuntimeError("Settings not loaded. Call load_settings() first.")
        try:
            return self._settings_cache[key_path]
        except (KeyError, TypeError):
            return default

    @staticmethod
    def parse_settings_files_from_string(content: str) -> list[str]:
        """
        Parse settings file from comma separated string.
        """
        if content.startswith("[") and content.endswith("]"):
            return json.loads(content)
        return [x.strip() for x in content.split(",") if x.strip()]



from typing import Any, Literal, Self

from pydantic import Field, PrivateAttr, model_validator

from gen_epix.filter.base import Filter
from gen_epix.filter.enum import FilterType


class StringSetFilter(Filter):
    members: frozenset[str] = Field(
        default=None, description="The strings to match.", frozen=True
    )
    case_sensitive: bool = Field(
        default=False, description="Whether the match is case sensitive.", frozen=True
    )
    _members: frozenset[str] = PrivateAttr()

    @model_validator(mode="after")
    def _validate_state(self) -> Self:
        if not self.case_sensitive:
            self._members = frozenset({x.lower() for x in self.members})
        else:
            self._members = self.members
        # Generate the function to check if a value is in the set of terms
        # The function is generated instead of defined to be able to optimize the check
        if self.case_sensitive:
            self._match = lambda x: x in self._members  # type: ignore
        else:
            self._match = lambda x: x.lower() in self._members  # type: ignore
        return self

    def _match(self, value: Any) -> bool:
        # Function is implemented dynamically in _validate_state
        raise NotImplementedError()


class TypedStringSetFilter(StringSetFilter):
    type: Literal[FilterType.STRING_SET.value]



# pylint: disable=too-few-public-methods


from typing import Any
from uuid import UUID

import sqlalchemy.orm as orm
from sqlalchemy.orm import Mapped

from gen_epix.casedb.domain import DOMAIN, enum, model
from gen_epix.commondb.repositories.sa_model import (
    RowMetadataMixin,
    create_mapped_column,
    create_table_args,
)

Base: type = orm.declarative_base(name=enum.ServiceType.SUBJECT.value)


class Subject(Base, RowMetadataMixin):
    __tablename__, __table_args__ = create_table_args(model.Subject)

    data_collection_id: Mapped[UUID] = create_mapped_column(
        DOMAIN, model.Subject, "data_collection_id"
    )
    external_ids: Mapped[dict[UUID, str] | None] = create_mapped_column(
        DOMAIN, model.Subject, "external_ids"
    )
    content: Mapped[dict[str, Any]] = create_mapped_column(
        DOMAIN, model.Subject, "content"
    )


class SubjectIdentifier(Base, RowMetadataMixin):
    __tablename__, __table_args__ = create_table_args(model.SubjectIdentifier)

    subject_id: Mapped[UUID] = create_mapped_column(
        DOMAIN, model.SubjectIdentifier, "subject_id"
    )
    identifier_issuer_id: Mapped[UUID] = create_mapped_column(
        DOMAIN, model.SubjectIdentifier, "identifier_issuer_id"
    )
    identifier: Mapped[str] = create_mapped_column(
        DOMAIN, model.SubjectIdentifier, "identifier"
    )



from typing import Literal

from pydantic import Field

from gen_epix.filter.enum import FilterType
from gen_epix.filter.equals import EqualsFilter


class EqualsStringFilter(EqualsFilter):
    value: str = Field(description="The string to match.", frozen=True)


class TypedEqualsStringFilter(EqualsStringFilter):
    type: Literal[FilterType.EQUALS_STRING.value]



from gen_epix.casedb.domain.repository import BaseOntologyRepository
from gen_epix.fastapp.repositories import SARepository


class OntologySARepository(SARepository, BaseOntologyRepository):
    pass



import abc
from collections.abc import Callable, Hashable, Iterable
from typing import Any

from sqlalchemy import Row
from sqlalchemy.orm import MappedColumn

from gen_epix.fastapp import exc
from gen_epix.fastapp.enum import FieldType, FieldTypeSet
from gen_epix.fastapp.model import Model


class BaseSAMapper(abc.ABC):
    def __init__(
        self,
        model_class: type[Model],
        row_class: type[Row],
        **kwargs: Any,
    ):
        if model_class.ENTITY is None:
            raise exc.RepositoryServiceError(
                f"Model {model_class.__name__} does not have an ENTITY class attribute"
            )
        self._model_class = model_class
        self._row_class = row_class
        self._table_name: str = row_class.__tablename__  # type: ignore[attr-defined]
        self._schema_name: str | None = self._get_schema_name(row_class)

    @property
    def model_class(self) -> type[Model]:
        return self._model_class

    @property
    def row_class(self) -> type:
        return self._row_class

    @property
    def table_name(self) -> str:
        return self._table_name

    @property
    def schema_name(self) -> str | None:
        return self._schema_name

    @abc.abstractmethod
    def get_field_names_by_type(self, field_type: FieldType) -> tuple:
        raise NotImplementedError()

    @abc.abstractmethod
    def get_field_names_by_set(self, field_type_set: FieldTypeSet) -> tuple:
        raise NotImplementedError()

    def get_field_name_map(self, reverse: bool = False) -> dict[str, str]:
        """
        Get a field name map between model and row fields. If one of the fields does
        not exist or there is no one-to-one mapping, it is excluded from the map.
        This is a naieve implementation that can be overridden for performance.
        """
        return {
            field_name if not reverse else row_field_name: (
                row_field_name if not reverse else field_name
            )
            for field_type in FieldType
            for field_name, row_field_name in zip(
                self.get_field_names_by_type(field_type),
                self.get_row_field_names_by_type(field_type),
            )
            if field_name and row_field_name
        }

    @abc.abstractmethod
    def get_mapped_field_name(
        self, field_name: str, reverse: bool = False
    ) -> str | None:
        """
        Get the mapped field name between model and row fields. If there is no
        equivalent or no one-to-one mapping, return None.
        This is a naieve implementation that can be overridden for performance.
        """
        return self.get_field_name_map(reverse).get(field_name)

    @abc.abstractmethod
    def get_row_field_names_by_type(self, field_type: FieldType) -> tuple:
        raise NotImplementedError()

    @abc.abstractmethod
    def get_row_field_names_by_set(self, field_type_set: FieldTypeSet) -> tuple:
        raise NotImplementedError()

    @abc.abstractmethod
    def get_id(self, obj: Model) -> Hashable | MappedColumn:
        raise NotImplementedError()

    @abc.abstractmethod
    def get_row_id(self, row: Row | type[Row]) -> Hashable:
        raise NotImplementedError()

    def generate_service_metadata(self, obj: Model, user_id: Hashable) -> dict:
        raise NotImplementedError()

    @abc.abstractmethod
    def dump(self, user_id: Hashable | None, obj: Model, **kwargs: Any) -> Any:
        raise NotImplementedError

    @abc.abstractmethod
    def load(self, row: Row, **kwargs: Any) -> Model:
        raise NotImplementedError

    @staticmethod
    def _get_schema_name(row: Row | type[Row]) -> str | None:
        for arg in row.__table_args__:  # type: ignore[union-attr]
            if isinstance(arg, dict) and "schema" in arg:
                return arg["schema"]  # type: ignore[no-any-return]
        return None


class SAMapper(BaseSAMapper):

    def __init__(
        self,
        model_class: type[Model],
        row_class: type,
        field_name_map: dict[str, str] | None = None,
        service_metadata_field_names: tuple | None = None,
        db_metadata_field_names: tuple | None = None,
        generate_service_metadata: (
            Callable[[Model, Hashable], dict[str, Any]] | None
        ) = None,
        **kwargs: Any,
    ):
        super().__init__(model_class, row_class, **kwargs)
        self.field_name_map = field_name_map or {}
        self.rev_field_name_map = {y: x for x, y in self.field_name_map.items()}
        self._field_names_by_type: dict[FieldType, tuple] = {}
        self._row_field_names_by_type: dict[FieldType, tuple] = {}
        self._field_names_by_set: dict[FieldTypeSet, tuple] = {}
        self._row_field_names_by_set: dict[FieldTypeSet, tuple] = {}
        self._relationship_field_name_map: dict[str, str] = {}
        self._relationship_field_name_reverse_map: dict[str, str] = {}
        self._init_field_names(model_class, row_class, self.field_name_map)
        self._init_row_metadata_field_names(
            row_class, service_metadata_field_names, db_metadata_field_names
        )
        self._init_relationship_field_names(model_class, row_class, self.field_name_map)
        self._init_extract_primary_key(model_class)
        self._generate_service_metadata = (
            generate_service_metadata if generate_service_metadata else lambda x, y: {}
        )
        self._is_identical_common_field_names = (
            self._field_names_by_set[FieldTypeSet.MODEL_DB_COMMON]
            == self._row_field_names_by_set[FieldTypeSet.MODEL_DB_COMMON]
        )
        self.field_name_map = {
            x: y
            for x, y in zip(
                self._field_names_by_set[FieldTypeSet.MODEL_DB_COMMON],
                self._row_field_names_by_set[FieldTypeSet.MODEL_DB_COMMON],
            )
        }
        self.rev_field_name_map = {y: x for x, y in self.field_name_map.items()}

    def get_field_names_by_type(self, field_type: FieldType) -> tuple:
        """
        The order of field names is guaranteed to be the same as the order of the row
        field names returned by the corresponding function.
        """
        return self._field_names_by_type[field_type]

    def get_field_names_by_set(self, field_type_set: FieldTypeSet) -> tuple:
        """
        The order of field names is guaranteed to be the same as the order of the row
        field names returned by the corresponding function.
        """
        return self._field_names_by_set[field_type_set]

    def get_field_name_map(self, reverse: bool = False) -> dict[str, str]:
        if reverse:
            return self.rev_field_name_map
        return self.field_name_map

    def get_mapped_field_name(
        self, field_name: str, reverse: bool = False
    ) -> str | None:
        if self._is_identical_common_field_names:
            return field_name
        if reverse:
            return self.rev_field_name_map.get(field_name)
        return self.field_name_map.get(field_name)

    def get_row_field_names_by_type(self, field_type: FieldType) -> tuple:
        return self._row_field_names_by_type[field_type]

    def get_row_field_names_by_set(self, field_type_set: FieldTypeSet) -> tuple:
        return self._row_field_names_by_set[field_type_set]

    def get_id(self, obj: Model) -> Hashable:
        return self._get_id(obj)

    def get_row_id(self, row: Row | type[Row]) -> Hashable | MappedColumn:
        return self._get_row_id(row)

    def generate_service_metadata(self, obj: Model, user_id: Hashable) -> dict:
        return self._generate_service_metadata(obj, user_id)

    def dump(self, user_id: Hashable | None, obj: Model, **kwargs: Any) -> Any:
        service_metadata = self.generate_service_metadata(obj, user_id)
        if self._is_identical_common_field_names:
            mapped_dict = obj.model_dump(exclude_none=True)
        else:
            obj_dict = obj.model_dump(exclude_none=False)
            mapped_dict = {
                y: obj_dict[x]
                for x, y in zip(
                    self._field_names_by_set[FieldTypeSet.MODEL_DB_COMMON],
                    self._row_field_names_by_set[FieldTypeSet.MODEL_DB_COMMON],
                )
                if obj_dict[x] is not None
            }
        if service_metadata:
            if kwargs:
                return self.row_class(**(mapped_dict | service_metadata | kwargs))
            return self.row_class(**(mapped_dict | service_metadata))
        if kwargs:
            return self.row_class(**(mapped_dict | kwargs))
        return self.row_class(**mapped_dict)

    def load(self, row: Row, **kwargs: Any) -> Model:
        if self._is_identical_common_field_names:
            mapped_dict = {
                x: getattr(row, x)
                for x in self._row_field_names_by_set[FieldTypeSet.MODEL_DB_COMMON]
            }
        else:
            mapped_dict = {
                x: getattr(row, y)
                for x, y in zip(
                    self._field_names_by_set[FieldTypeSet.MODEL_DB_COMMON],
                    self._row_field_names_by_set[FieldTypeSet.MODEL_DB_COMMON],
                )
            }
        if kwargs:
            return self.model_class(**(mapped_dict | kwargs))
        return self.model_class(**mapped_dict)

    def _init_field_names(
        self, model_class: type[Model], row_class: type, field_name_map: dict[str, str]
    ) -> None:
        # Set model and row field names by field type
        valid_row_field_names = set(row_class.__table__.columns.keys())
        entity = model_class.ENTITY
        if entity is None:
            raise exc.RepositoryInitializationServiceError(
                f"Model {model_class.__name__} does not ENTITY set"
            )
        common_field_names_set = set.union(
            *[
                set(entity.get_field_names(field_type=x))
                for x in FieldTypeSet.MODEL_DB_COMMON.value
            ],
        )
        for field_type in FieldType:
            field_names = entity.get_field_names(field_type=field_type)
            row_field_names = [
                field_name_map.get(x, x)
                for x in field_names
                if x in common_field_names_set
            ]
            invalid_row_field_names = set(row_field_names) - valid_row_field_names
            if invalid_row_field_names:
                invalid_row_field_names_str = ", ".join(invalid_row_field_names)
                raise exc.RepositoryServiceError(
                    f"Row {row_class.__name__} field(s) {invalid_row_field_names_str} do not exist"
                )
            self._field_names_by_type[field_type] = tuple(field_names)
            self._row_field_names_by_type[field_type] = tuple(row_field_names)

        # Set model and row field names by field type set
        for field_type_set in FieldTypeSet:
            field_names = []
            row_field_names = []
            for field_type in field_type_set.value:
                field_names.extend(self._field_names_by_type[field_type])
                row_field_names.extend(self._row_field_names_by_type[field_type])
            self._field_names_by_set[field_type_set] = tuple(field_names)
            self._row_field_names_by_set[field_type_set] = tuple(row_field_names)

    def _init_row_metadata_field_names(
        self,
        row_class: type,
        service_metadata_field_names: Iterable[str] | None,
        db_metadata_field_names: Iterable[str] | None,
    ) -> None:

        # Helper function to check provided field names
        def _check_field_names(
            field_names: Iterable[str] | None,
            name: str,
            valid_field_names: Iterable[str],
        ) -> tuple[str, ...] | None:
            if field_names is None:
                return None
            if isinstance(field_names, Iterable):
                out_field_names: tuple[str, ...] = tuple(field_names)
            else:
                raise exc.RepositoryServiceError(
                    f"Row {row_class.__name__} {name} must be a tuple"
                )
            invalid_field_names = set(out_field_names) - set(valid_field_names)
            if invalid_field_names:
                invalid_field_names_str = ", ".join(invalid_field_names)
                raise exc.RepositoryServiceError(
                    f"Row {row_class.__name__} {name} provided field(s) {invalid_field_names_str} are not in the db model"
                )
            return out_field_names

        # Check provided field names
        row_field_names = set(
            self._row_field_names_by_set[FieldTypeSet.MODEL_DB_COMMON]
        )
        actual_row_only_field_names = [
            x for x in row_class.__table__.columns.keys() if x not in row_field_names
        ]
        actual_row_only_field_names_set = set(actual_row_only_field_names)
        service_metadata_field_names = _check_field_names(
            service_metadata_field_names,
            "service_metadata_field_names",
            actual_row_only_field_names_set,
        )
        db_metadata_field_names = _check_field_names(
            db_metadata_field_names,
            "db_metadata_field_names",
            actual_row_only_field_names_set,
        )

        # Set service_metadata_field_names and db_metadata_field_names
        if service_metadata_field_names is None:
            if db_metadata_field_names is None:
                # All db model only fields are considered db_metadata_field_names
                service_metadata_field_names = tuple()
                service_metadata_field_names_set = set()
                db_metadata_field_names = tuple(actual_row_only_field_names)
                db_metadata_field_names_set = set(db_metadata_field_names)
            else:
                # service_metadata_field_names is the complement of db_metadata_field_names
                db_metadata_field_names_set = set(db_metadata_field_names)
                service_metadata_field_names = tuple(
                    x
                    for x in actual_row_only_field_names
                    if x not in db_metadata_field_names_set
                )
                service_metadata_field_names_set = set(service_metadata_field_names)
        elif db_metadata_field_names is None:
            # db_metadata_field_names is the complement of service_metadata_field_names
            service_metadata_field_names_set = set(service_metadata_field_names)
            db_metadata_field_names = tuple(
                x
                for x in actual_row_only_field_names
                if x not in service_metadata_field_names_set
            )
            db_metadata_field_names_set = set(db_metadata_field_names)
        else:
            # Both are provided, create sets for validation
            service_metadata_field_names_set = set(service_metadata_field_names)
            db_metadata_field_names_set = set(db_metadata_field_names)

        # Final check of field names
        if not service_metadata_field_names_set.isdisjoint(db_metadata_field_names_set):
            raise exc.RepositoryServiceError(
                f"Row {row_class.__name__} service_metadata_field_names and db_metadata_field_names must be disjoint"
            )
        metadata_field_names_set = service_metadata_field_names_set.union(
            db_metadata_field_names_set
        )
        if metadata_field_names_set != actual_row_only_field_names_set:
            raise exc.RepositoryServiceError(
                f"Row {row_class.__name__} service_metadata_field_names and db_metadata_field_names must cover all db model only fields"
            )

        # Update attributes
        self._row_field_names_by_type[FieldType.SERVICE_METADATA] = (
            service_metadata_field_names
        )
        self._row_field_names_by_type[FieldType.DB_METADATA] = db_metadata_field_names
        for field_type_set in {
            FieldTypeSet.SERVICE_METADATA,
            FieldTypeSet.DB_METADATA,
            FieldTypeSet.DB_MODEL,
            FieldTypeSet.DB_MODEL_ONLY,
        }:
            self._row_field_names_by_set[field_type_set] = tuple(
                self._row_field_names_by_set[field_type_set]
            )

    def _init_relationship_field_names(
        self, model_class: type[Model], row_class: type, field_name_map: dict[str, str]
    ) -> None:
        # Check that all link fields in the model have corresponding relationship fields in the row
        entity = model_class.ENTITY
        if entity is None:
            raise exc.RepositoryInitializationServiceError(
                f"Model {model_class.__name__} does not have an ENTITY set"
            )
        relationship_field_names = entity.get_field_names(
            field_type=FieldType.RELATIONSHIP
        )
        row_relationship_field_names = [
            field_name_map.get(x, x) for x in relationship_field_names
        ]
        for field_name, row_field_name in zip(
            relationship_field_names, row_relationship_field_names
        ):
            if not hasattr(row_class, row_field_name):
                continue
            self._relationship_field_name_map[field_name] = row_field_name
            self._relationship_field_name_reverse_map[row_field_name] = field_name

    def _init_extract_primary_key(self, model_class: type[Model]) -> None:
        id_field_names = self._field_names_by_type[FieldType.ID]
        row_id_field_names = self._row_field_names_by_type[FieldType.ID]
        if len(id_field_names) != 1 or len(row_id_field_names) != 1:
            raise NotImplementedError(
                f"Model {model_class.__name__} has more than one ID field"
            )
        id_field_name = id_field_names[0]
        row_id_field_name = row_id_field_names[0]
        self._get_id: Callable[[Model], Hashable] = lambda x: getattr(x, id_field_name)
        self._get_row_id: Callable[[Row | type[Row]], Hashable | MappedColumn] = (
            lambda x: getattr(x, row_id_field_name)
        )



import abc
import uuid
from typing import Any

from gen_epix.commondb.domain import command, model
from gen_epix.commondb.domain.enum import ServiceType
from gen_epix.commondb.domain.repository.organization import BaseOrganizationRepository
from gen_epix.fastapp import BaseService
from gen_epix.fastapp.model import UpdateAssociationCommand


class BaseOrganizationService(BaseService):
    SERVICE_TYPE = ServiceType.ORGANIZATION

    # Property overridden to provide narrower return value to support linter
    @property  # type: ignore
    def repository(self) -> BaseOrganizationRepository:  # type: ignore
        return super().repository  # type: ignore

    def register_handlers(self) -> None:
        f = self.app.register_handler
        self.register_default_crud_handlers()
        for command_class in self.app.domain.get_commands_for_service_type(
            self.service_type, base_class=UpdateAssociationCommand
        ):
            f(command_class, self.update_association)
        f(
            command.RetrieveOrganizationContactCommand,
            self.retrieve_organization_contact,
        )
        f(command.InviteUserCommand, self.invite_user)
        f(
            command.RetrieveInviteUserConstraintsCommand,
            self.retrieve_invite_user_constraints,
        )
        f(command.RegisterInvitedUserCommand, self.register_invited_user)
        f(command.UpdateUserCommand, self.update_user)

    @abc.abstractmethod
    def retrieve_organization_contact(
        self,
        cmd: command.RetrieveOrganizationContactCommand,
    ) -> list[model.Contact]:
        raise NotImplementedError()

    @abc.abstractmethod
    def retrieve_user_by_key(self, user_key: str) -> model.User:
        raise NotImplementedError()

    @abc.abstractmethod
    def invite_user(
        self,
        cmd: command.InviteUserCommand,
    ) -> model.UserInvitation:
        raise NotImplementedError

    @abc.abstractmethod
    def retrieve_invite_user_constraints(
        self, cmd: command.RetrieveInviteUserConstraintsCommand
    ) -> model.UserInvitationConstraints:
        raise NotImplementedError()

    @abc.abstractmethod
    def register_invited_user(
        self, cmd: command.RegisterInvitedUserCommand
    ) -> model.User:
        raise NotImplementedError

    def generate_user_invitation_token(self, **kwargs: Any) -> str:
        return str(uuid.uuid4())
        return str(uuid.uuid4())

    @abc.abstractmethod
    def update_user(
        self,
        cmd: command.UpdateUserCommand,
    ) -> model.User:
        raise NotImplementedError



from typing import Any

from gen_epix.commondb.services import OrganizationService as CommonOrganizationService
from gen_epix.seqdb.domain import model


class OrganizationService(CommonOrganizationService):
    def __init__(
        self,
        *args: Any,
        **kwargs: Any,
    ) -> None:
        super().__init__(
            *args,
            **kwargs,
            user_class=model.User,
            user_invitation_class=model.UserInvitation,
            user_invitation_constraints_class=model.UserInvitationConstraints,
        )



import datetime
import ipaddress
from decimal import Decimal
from enum import Enum
from pathlib import Path
from typing import Any
from uuid import UUID

import sqlalchemy as sa
from pydantic import BaseModel
from pydantic.fields import FieldInfo
from sqlalchemy.ext.compiler import compiles
from sqlalchemy.sql import expression
from sqlalchemy.sql.compiler import SQLCompiler
from sqlalchemy.types import DateTime, TypeEngine
from sqlalchemy_utils.types.uuid import UUIDType

from gen_epix.fastapp.domain.util import get_type_from_annotation

PYTHON_SQL_TYPE_MAP = {
    str: sa.String,  # sa.Text, sa.Unicode, sa.UnicodeText can be chosen
    int: sa.Integer,
    float: sa.Float,
    bool: sa.Boolean,
    datetime.datetime: sa.DateTime,
    datetime.date: sa.Date,
    datetime.time: sa.Time,
    datetime.timedelta: sa.Interval,
    Decimal: sa.Numeric,
    bytes: sa.LargeBinary,
    UUID: UUIDType,
    ipaddress.IPv4Address: sa.String,
    ipaddress.IPv6Address: sa.String,
    ipaddress.IPv4Network: sa.String,
    ipaddress.IPv6Network: sa.String,
    Path: sa.String,
    dict: sa.JSON,
    list: sa.JSON,
    set: sa.JSON,
    frozenset: sa.JSON,
    tuple: sa.JSON,
}

PYDANTIC_SA_FIELD_METADATA_MAP: dict[str, str] = {
    "max_length": "length",
    "max_digits": "precision",
    "decimal_places": "scale",
}

SA_METADATA_BY_TYPE: dict[type[TypeEngine], frozenset[str]] = {
    sa.String: frozenset({"length", "collation"}),
    sa.Unicode: frozenset({"length", "collation"}),
    sa.Text: frozenset({"collation"}),
    sa.UnicodeText: frozenset({"collation"}),
    sa.Boolean: frozenset({"create_constraint", "name"}),
    sa.Integer: frozenset({}),
    sa.BigInteger: frozenset({}),
    sa.SmallInteger: frozenset({}),
    sa.Float: frozenset({"precision", "asdecimal", "decimal_return_scale"}),
    sa.Numeric: frozenset({"precision", "scale", "decimal_return_scale", "asdecimal"}),
    sa.DECIMAL: frozenset({"precision", "scale", "decimal_return_scale", "asdecimal"}),
    sa.DateTime: frozenset({"timezone"}),
    sa.Date: frozenset({}),
    sa.Time: frozenset({"timezone"}),
    sa.Interval: frozenset({"native", "second_precision", "day_precision"}),
    sa.LargeBinary: frozenset({"length"}),
    sa.BINARY: frozenset({"length"}),
    sa.VARBINARY: frozenset({"length"}),
    sa.JSON: frozenset({"none_as_null"}),
    sa.CHAR: frozenset({"length", "collation"}),
    sa.VARCHAR: frozenset({"length", "collation"}),
    sa.NCHAR: frozenset({"length", "collation"}),
    sa.NVARCHAR: frozenset({"length", "collation"}),
    sa.CLOB: frozenset({"collation"}),
    sa.BLOB: frozenset({"length"}),
    sa.TIMESTAMP: frozenset({"timezone"}),
    UUIDType: frozenset({"binary", "native"}),
}


class ServerUtcTimestamp(expression.FunctionElement):
    type = sa.TIMESTAMP()
    inherit_cache = True


@compiles(ServerUtcTimestamp, "postgresql")
def postgresql_utc_timestamp(
    _element: ServerUtcTimestamp, _compiler: SQLCompiler, **_kw: dict
) -> str:
    return "TIMEZONE('utc', CURRENT_TIMESTAMP)"


@compiles(ServerUtcTimestamp, "mssql")
def mssql_utc_timestamp(
    _element: ServerUtcTimestamp, _compiler: SQLCompiler, **_kw: dict
) -> str:
    return "GETUTCDATE()"


@compiles(ServerUtcTimestamp, "sqlite")
def sqlite_utc_timestamp(
    _element: ServerUtcTimestamp, _compiler: SQLCompiler, **_kw: dict
) -> str:
    return "STRFTIME('%Y-%m-%d %H:%M:%f', 'NOW')"


class ServerUtcCurrentTime(expression.FunctionElement):
    type = DateTime()
    inherit_cache = True


@compiles(ServerUtcCurrentTime, "postgresql")
def postgresql_utc_current_time(
    _element: ServerUtcCurrentTime, _compiler: SQLCompiler, **_kw: dict
) -> str:
    return "TIMEZONE('utc', CURRENT_TIMESTAMP)"


@compiles(ServerUtcCurrentTime, "mssql")
def mssql_utc_current_time(
    _element: ServerUtcCurrentTime, _compiler: SQLCompiler, **_kw: dict
) -> str:
    return "GETUTCDATE()"


@compiles(ServerUtcCurrentTime, "sqlite")
def sqlite_utc_current_time(
    _element: ServerUtcCurrentTime, _compiler: SQLCompiler, **_kw: dict
) -> str:
    return "CURRENT_TIMESTAMP"


def create_sa_type_from_field_info(
    field_info: FieldInfo, annotation: type[Any] | None, **kwargs: dict
) -> TypeEngine:
    """
    Return a suitable SQLAlchemy type for a Pydantic field.
    """
    type_ = get_type_from_annotation(annotation)

    def _create_sa_type(sa_type_class: type[TypeEngine]) -> TypeEngine:
        # Get column kwargs for this type, overridden by kwargs
        new_kwargs = (
            get_sa_type_kwargs_from_field_info(sa_type_class, field_info) | kwargs
        )
        # Special case: String without length becomes Text
        if sa_type_class is sa.String and "length" not in new_kwargs:
            sa_type_class = sa.Text
            new_kwargs = (
                get_sa_type_kwargs_from_field_info(sa_type_class, field_info) | kwargs
            )
        # Special case: Unicode without length becomes UnicodeText
        if sa_type_class is sa.Unicode and "length" not in new_kwargs:
            sa_type_class = sa.UnicodeText
            new_kwargs = (
                get_sa_type_kwargs_from_field_info(sa_type_class, field_info) | kwargs
            )
        return sa_type_class(**new_kwargs)

    if issubclass(type_, Enum):
        # Special case: construct from type itself
        return sa.Enum(type_)
    if type_ in PYTHON_SQL_TYPE_MAP:
        return _create_sa_type(PYTHON_SQL_TYPE_MAP[type_])
    if issubclass(type_, BaseModel):
        # Special case: pydantic models as JSON
        return _create_sa_type(sa.JSON)

    raise NotImplementedError(f"Unsupported field type: {type_}")


def get_sa_type_kwargs_from_field_info(
    sa_type_class: type[sa.types.TypeEngine], field_info: FieldInfo
) -> dict[str, Any]:
    # Extract column kwargs from field metadata
    kwargs: dict[str, Any] = {}
    for metadata in field_info.metadata:
        for pydantic_name, sa_name in PYDANTIC_SA_FIELD_METADATA_MAP.items():
            if hasattr(metadata, pydantic_name):
                kwargs[sa_name] = getattr(metadata, pydantic_name)
    # Restrict column kwargs to allowed ones for this particular column type
    if sa_type_class not in SA_METADATA_BY_TYPE:
        raise NotImplementedError(
            f"Unsupported SQLAlchemy column type: {sa_type_class}"
        )
    return {x: y for x, y in kwargs.items() if x in SA_METADATA_BY_TYPE[sa_type_class]}



from gen_epix.commondb.services import AbacService as CommonAbacService
from gen_epix.fastapp.model import Command
from gen_epix.omopdb.domain import command


class BaseAbacService(CommonAbacService):
    ORGANIZATION_ADMIN_WRITE_COMMANDS: set[type[Command]] = {  # type: ignore[assignment]
        command.COMMON_COMMAND_MAP.get(x, x)
        for x in CommonAbacService.ORGANIZATION_ADMIN_WRITE_COMMANDS
    } | set()

    READ_ORGANIZATION_RESULTS_ONLY_COMMANDS: set[type[Command]] = {  # type: ignore[assignment]
        command.COMMON_COMMAND_MAP.get(x, x)
        for x in CommonAbacService.READ_ORGANIZATION_RESULTS_ONLY_COMMANDS
    } | set()

    READ_SELF_RESULTS_ONLY_COMMANDS: set[type[Command]] = {  # type: ignore[assignment]
        command.COMMON_COMMAND_MAP.get(x, x)
        for x in CommonAbacService.READ_SELF_RESULTS_ONLY_COMMANDS
    } | set()

    READ_USER_COMMANDS: set[type[Command]] = {  # type: ignore[assignment]
        command.COMMON_COMMAND_MAP.get(x, x)
        for x in CommonAbacService.READ_USER_COMMANDS
    } | set()

    UPDATE_USER_COMMANDS: set[type[Command]] = {  # type: ignore[assignment]
        command.COMMON_COMMAND_MAP.get(x, x)
        for x in CommonAbacService.UPDATE_USER_COMMANDS
    } | set()



# pylint: disable=useless-import-alias

from gen_epix.casedb.domain.service.abac import BaseAbacService as BaseAbacService
from gen_epix.casedb.domain.service.case import BaseCaseService as BaseCaseService
from gen_epix.casedb.domain.service.geo import BaseGeoService as BaseGeoService
from gen_epix.casedb.domain.service.ontology import (
    BaseOntologyService as BaseOntologyService,
)
from gen_epix.casedb.domain.service.seqdb import BaseSeqdbService as BaseSeqdbService
from gen_epix.casedb.domain.service.subject import (
    BaseSubjectService as BaseSubjectService,
)
from gen_epix.commondb.domain.service import (
    BaseOrganizationService as BaseOrganizationService,
)
from gen_epix.commondb.domain.service import BaseRbacService as BaseRbacService
from gen_epix.commondb.domain.service import BaseSystemService as BaseSystemService
from gen_epix.fastapp.services.auth import BaseAuthService as BaseAuthService



from typing import Any, ClassVar
from uuid import UUID

from pydantic import Field, model_validator

import gen_epix.commondb.domain.model.organization as model
from gen_epix.commondb.domain.command.base import (
    Command,
    CrudCommand,
    UpdateAssociationCommand,
)
from gen_epix.commondb.util import copy_model_field

# Non-CRUD commands


class OrganizationSetOrganizationUpdateAssociationCommand(UpdateAssociationCommand):
    """
    Updates the association between an {organization_set}s and {organization}s.

    This command manages the many-to-many relationship by creating or updating
    {organization_set_member} associations between organization sets and
    individual organizations.
    """

    __doc__ = str(__doc__).format(
        organization_set=model.OrganizationSet.__name__,
        organization=model.Organization.__name__,
        organization_set_member=model.OrganizationSetMember.__name__,
    )

    ASSOCIATION_CLASS: ClassVar = model.OrganizationSetMember
    LINK_FIELD_NAME1: ClassVar = "organization_set_id"
    LINK_FIELD_NAME2: ClassVar = "organization_id"

    obj_id1: UUID | None = None
    obj_id2: UUID | None = None
    association_objs: list[model.OrganizationSetMember]


class DataCollectionSetDataCollectionUpdateAssociationCommand(UpdateAssociationCommand):
    """
    Updates the association between {data_collection_set}s and {data_collection}s.

    This command manages the many-to-many relationship by creating or updating
    {data_collection_set_member} associations between data collection sets and
    individual data collections.
    """

    __doc__ = str(__doc__).format(
        data_collection_set=model.DataCollectionSet.__name__,
        data_collection=model.DataCollection.__name__,
        data_collection_set_member=model.DataCollectionSetMember.__name__,
    )

    ASSOCIATION_CLASS: ClassVar = model.DataCollectionSetMember
    LINK_FIELD_NAME1: ClassVar = "data_collection_set_id"
    LINK_FIELD_NAME2: ClassVar = "data_collection_id"

    obj_id1: UUID | None = None
    obj_id2: UUID | None = None
    association_objs: list[model.DataCollectionSetMember]


class InviteUserCommand(Command):
    """
    Creates and returns a {user_invitation} for a new user with a particular
    user key claim (e.g. email address), organization and initial role(s).

    A random unique token is added to the invitation, and to be provided to the
    new user for consuming the invitation.
    """

    __doc__ = str(__doc__).format(user_invitation=model.UserInvitation.__name__)

    key: str = copy_model_field(model.UserInvitation, "key")
    email: str | None = copy_model_field(model.UserInvitation, "email")
    name: str | None = copy_model_field(model.UserInvitation, "name")
    roles: set[str] = copy_model_field(model.UserInvitation, "roles")
    organization_id: UUID = copy_model_field(model.UserInvitation, "organization_id")


class RegisterInvitedUserCommand(Command):
    """
    Registers (creates) the user of the command. The key and token must match
    that of an existing {user_invitation}. The newly registered user is assigned
    the organization and roles from the invitation. The invitation is deleted.
    """

    __doc__ = str(__doc__).format(user_invitation=model.UserInvitation.__name__)

    token: str = copy_model_field(model.UserInvitation, "token")


class RetrieveOrganizationContactCommand(Command):
    """
    Retrieves {contact}s associated with organizations, sites, or specific contacts.

    Exactly one of organization_ids, site_ids, or contact_ids must be provided.
    Returns a list of contacts with their associated site and organization data
    cascaded.
    """

    __doc__ = str(__doc__).format(contact=model.Contact.__name__)

    organization_ids: list[UUID] | None = Field(
        default=None, description="List of organization IDs to retrieve contacts for"
    )
    site_ids: list[UUID] | None = Field(
        default=None, description="List of site IDs to retrieve contacts for"
    )
    contact_ids: list[UUID] | None = Field(
        default=None, description="List of contact IDs to retrieve contacts for"
    )

    @model_validator(mode="after")
    def check_one_of_fields(self) -> Any:
        if (
            sum(
                [
                    self.organization_ids is not None,
                    self.site_ids is not None,
                    self.contact_ids is not None,
                ]
            )
            != 1
        ):
            raise ValueError(
                "Exactly one of organization_ids, site_ids or contact_ids must be "
                "provided"
            )
        return self


class UpdateUserCommand(Command):
    """
    Updates an existing {user} with new properties such as active status,
    roles, and organization membership.

    The target user is identified by tgt_user_id. Any field set to None will
    leave that property unchanged. Roles cannot be set to an empty set.
    Cache is invalidated after successful update.
    """

    __doc__ = str(__doc__).format(user=model.User.__name__)

    tgt_user_id: UUID = Field(description="The ID of the user to update")
    is_active: bool | None = copy_model_field(model.User, "is_active")
    roles: set[str] | None = copy_model_field(model.User, "roles")
    organization_id: UUID | None = Field(
        description="The organization ID the user belongs to"
    )


class UpdateUserOwnOrganizationCommand(Command):
    """
    Updates the current user's {organization} membership.

    This command allows a user to change their own organization association.
    The is_new_user flag indicates whether this is part of a new user
    registration process.
    """

    __doc__ = str(__doc__).format(organization=model.Organization.__name__)

    organization_id: UUID
    is_new_user: bool = False


class RetrieveInviteUserConstraintsCommand(Command):
    """
    Retrieves the constraints for inviting a user, such as valid roles and organizations.

    This command is used to gather the necessary information for the user invitation process.
    """

    pass


class RetrieveOrganizationAdminNameEmailsCommand(Command):
    """
    Retrieve the names and email addresses of all organization admins for the user's
    organization.
    """

    pass


# CRUD commands


class OrganizationCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.Organization


class UserCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.User


class UserInvitationCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.UserInvitation


class OrganizationSetCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.OrganizationSet


class OrganizationSetMemberCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.OrganizationSetMember


class SiteCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.Site


class ContactCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.Contact


class IdentifierIssuerCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.IdentifierIssuer


class DataCollectionCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.DataCollection


class DataCollectionSetCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.DataCollectionSet


class DataCollectionSetMemberCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.DataCollectionSetMember



# pylint: disable=too-few-public-methods
# This module defines base classes, methods are added later


import datetime
from typing import ClassVar

from pydantic import Field

from gen_epix.commondb.domain import enum
from gen_epix.commondb.domain.model.base import Model
from gen_epix.fastapp import Entity


class Outage(Model):
    """
    Represents a system outage.
    """

    ENTITY: ClassVar = Entity(
        snake_case_plural_name="outages",
        table_name="outage",
        persistable=True,
    )
    description: str | None = Field(
        default=None, description="Description of the system outage."
    )
    active_from: datetime.datetime | None = Field(
        default=None, description="The date-time when the system outage starts."
    )
    active_to: datetime.datetime | None = Field(
        default=None, description="The date-time when the system outage ends."
    )
    visible_from: datetime.datetime | None = Field(
        default=None, description="The date-time when the system outage is announced."
    )
    visible_to: datetime.datetime | None = Field(
        default=None,
        description="The date-time when the system outage is no longer announced.",
    )
    is_active: bool | None = Field(
        default=None,
        description="Whether the system outage is currently active, this overrides active_from and active_to.",
    )
    is_visible: bool | None = Field(
        default=None,
        description="Whether the system outage is currently visible, this overrides visible_from and visible_to.",
    )


class PackageMetadata(Model):
    ENTITY: ClassVar = Entity(
        snake_case_plural_name="package_metadatas",
        persistable=False,
    )

    name: str = Field(description="Name of the package.")
    version: str = Field(description="Version of the package.")
    license: str | None = Field(
        default=None, description="License information for the package."
    )
    homepage: str | None = Field(
        default=None, description="Homepage URL of the package."
    )



from uuid import UUID

from gen_epix.casedb.domain import command, model
from gen_epix.casedb.services.case.base import BaseCaseService
from gen_epix.fastapp import BaseUnitOfWork, CrudOperation
from gen_epix.filter import CompositeFilter, Filter, LogicalOperator, UuidSetFilter


def case_service_read_association_with_valid_ids(
    self: BaseCaseService,
    command_class: type[command.CrudCommand],
    field_name1: str,
    field_name2: str,
    valid_ids1: set[UUID] | frozenset[UUID] | None = None,
    valid_ids2: set[UUID] | frozenset[UUID] | None = None,
    match_all1: bool = False,
    match_all2: bool = False,
    return_type: str = "objects",
    uow: BaseUnitOfWork | None = None,
    user: model.User | None = None,
) -> list[model.Model] | list[UUID] | dict[UUID, set[UUID]]:
    # TODO: this can be a generic service/repository method (ids should be Hashable instead of UUID)
    # Parse arguments
    if return_type not in {"objects", "ids1", "ids2", "id_map12", "id_map21"}:
        raise ValueError(f"Invalid return_type: {return_type}")
    if match_all1 and match_all2:
        raise ValueError("match_all1 and match_all2 cannot both be True")
    id_map12 = return_type == "id_map12"
    id_map21 = return_type == "id_map21"
    if id_map12 and match_all1:
        raise ValueError("match_all1 must be False if id_map12 is True")
    if id_map21 and match_all2:
        raise ValueError("match_all2 must be False if id_map21 is True")
    if return_type == "ids1" and match_all1:
        raise ValueError("match_all1 must be False if return_type is ids1")
    if return_type == "ids2" and match_all2:
        raise ValueError("match_all2 must be False if return_type is ids2")
    # Create filter
    filter: Filter | None
    if valid_ids1 is not None:
        if not isinstance(valid_ids1, frozenset):
            valid_ids1 = frozenset(valid_ids1)
        if not valid_ids1:
            # Empty set of valid values -> no matches
            if return_type in {"id_map12", "id_map21"}:
                return dict()
            return []
        if valid_ids2 is not None:
            if not valid_ids2:
                # Empty set of valid values -> no matches
                if return_type in {"id_map12", "id_map21"}:
                    return dict()
                return []
            if not isinstance(valid_ids2, frozenset):
                valid_ids2 = frozenset(valid_ids2)
            filter = CompositeFilter(
                filters=[
                    UuidSetFilter(key=field_name1, members=valid_ids1),
                    UuidSetFilter(key=field_name2, members=valid_ids2),
                ],
                operator=LogicalOperator.AND,
            )
        else:
            if match_all2:
                raise ValueError("match_all2 must be False if valid_ids2 is None")
            if not isinstance(valid_ids1, frozenset):
                valid_ids2 = frozenset(valid_ids2)
            filter = UuidSetFilter(key=field_name1, members=valid_ids1)
    elif valid_ids2 is not None:
        if not valid_ids2:
            # Empty set of valid values -> no matches
            if return_type in {"id_map12", "id_map21"}:
                return dict()
            return []
        if match_all1:
            raise ValueError("match_all1 must be False if valid_ids1 is None")
        if not isinstance(valid_ids2, frozenset):
            valid_ids2 = frozenset(valid_ids2)
        filter = UuidSetFilter(key=field_name2, members=valid_ids2)
    else:
        if match_all1 or match_all2:
            raise ValueError(
                "match_all1 and match_all2 must be False if valid_ids1 and valid_ids2 are None"
            )
        filter = None
    # Query repository
    cmd = command_class(
        user=user, operation=CrudOperation.READ_ALL, query_filter=filter
    )
    objs: list[model.Model]
    if uow:
        objs = self.crud_repository(uow, cmd)  # type: ignore[assignment]
    else:
        with self.repository.uow() as uow:
            objs = self.crud_repository(uow, cmd)  # type: ignore[assignment]
    ids1 = [getattr(x, field_name1) for x in objs]
    ids2 = [getattr(x, field_name2) for x in objs]
    # Apply id_map12/id_map21 and match_all1/match_all2 if necessary
    if id_map12 or id_map21 or match_all1 or match_all2:
        id_map: dict[UUID, set[UUID]] = {}
        if id_map12 or match_all2:
            # Create dict[id1, set[id2]]
            for id1, id2 in zip(ids1, ids2):
                if id1 in id_map:
                    id_map[id1].add(id2)
                else:
                    id_map[id1] = {id2}
            if match_all2:
                # Keep only ids1 linked to all valid ids2
                id_map = {
                    x: y for x, y in id_map.items() if len(y) == len(valid_ids2)  # type: ignore[arg-type]
                }
                if id_map12:
                    return id_map
                elif return_type == "objects":
                    return [x for x, y in zip(objs, ids1) if y in id_map]
                elif return_type == "ids1":
                    return list(id_map.keys())
            elif id_map12:
                return id_map
            else:
                raise AssertionError("Unexpected case")
        elif id_map21 or match_all1:
            # Create dict[id2, set[id1]]
            for id1, id2 in zip(ids1, ids2):
                if id2 in id_map:
                    id_map[id2].add(id1)
                else:
                    id_map[id2] = {id1}
            if match_all1:
                # Keep only ids2 linked to all valid ids1
                id_map = {
                    x: y for x, y in id_map.items() if len(y) == len(valid_ids1)  # type: ignore[arg-type]
                }
                if id_map21:
                    return id_map
                elif return_type == "objects":
                    return [x for x, y in zip(objs, ids2) if y in id_map]
                elif return_type == "ids2":
                    return list(id_map.keys())
            elif id_map21:
                return id_map
            else:
                raise AssertionError("Unexpected case")
        else:
            raise AssertionError("Unexpected case")
    # Return objs or ids for remaining cases
    if return_type == "objects":
        return objs
    if return_type == "ids1":
        return ids1
    if return_type == "ids2":
        return ids2
    raise AssertionError(f"Unexpected return_type: {return_type}")



import abc
from collections.abc import Callable
from typing import Any
from uuid import UUID

from gen_epix.commondb.domain.command import Command
from gen_epix.commondb.domain.service import BaseAbacService
from gen_epix.fastapp.model import Policy


class BaseAbacPolicy(Policy):
    def __init__(self, abac_service: BaseAbacService, **kwargs: Any):
        super().__init__()
        self.abac_service = abac_service
        self.props = kwargs


class BaseIsOrganizationAdminPolicy(BaseAbacPolicy):

    @abc.abstractmethod
    def register_retrieve_organization_ids_handler(
        self,
        command_class: type[Command],
        handler: Callable[[Command], set[UUID]],
    ) -> None:
        raise NotImplementedError

    @abc.abstractmethod
    def retrieve_organization_ids(self, cmd: Command) -> set[UUID]:
        raise NotImplementedError


class BaseReadOrganizationResultsOnlyPolicy(BaseAbacPolicy):
    pass


class BaseReadSelfResultsOnlyPolicy(BaseAbacPolicy):
    pass


class BaseReadUserPolicy(BaseAbacPolicy):
    pass


class BaseUpdateUserPolicy(BaseAbacPolicy):
    pass



from collections.abc import Hashable

from pydantic import Field

from gen_epix.filter.base import Filter


class HashableSetFilter(Filter):
    members: frozenset[Hashable] = Field(
        default=None, description="The values to match.", frozen=True
    )

    def _match(self, value: Hashable) -> bool:
        return value in self.members


# No typed version of this filter is needed since the type of the values would be needed as well
# class TypedValueSetFilter(ValueSetFilter):
#     type_: Literal[FilterType.HASHABLE_SET]



"""
Stream processing interfaces and base classes.
"""

from abc import ABC, abstractmethod
from collections.abc import Iterator
from typing import Any

from gen_epix.transform.transform_result import TransformResult


class StreamProcessor(ABC):
    """Base interface for stream processing components."""

    @abstractmethod
    def process_stream(self, stream: Iterator[Any]) -> Iterator[TransformResult]:
        """
        Process a stream of objects and yield transformation results.

        Args:
            stream: Iterator of objects to process

        Yields:
            TransformResult objects containing success/failure information
        """
        pass



from typing import Any, Literal, Self

from pydantic import Field, model_validator

from gen_epix.filter.base import Filter
from gen_epix.filter.enum import ComparisonOperator, FilterType


class RangeFilter(Filter):
    lower_bound: Any | None = Field(
        default=None, description="The lower bound of the range.", frozen=True
    )
    lower_bound_censor: ComparisonOperator = Field(
        default=ComparisonOperator.GTE,
        description="The censor of the lower bound.",
        frozen=True,
    )
    upper_bound: Any | None = Field(
        default=None, description="The upper bound of the range.", frozen=True
    )
    upper_bound_censor: ComparisonOperator = Field(
        default=ComparisonOperator.ST,
        description="The censor of the upper bound.",
        frozen=True,
    )

    def _validate_state_bounds(self) -> None:
        # Validate the bounds and censors
        if self.lower_bound is None:
            if self.upper_bound is None:
                raise AssertionError("At least one bound must be set.")
        else:
            if self.upper_bound is not None:
                if self.lower_bound > self.upper_bound:
                    raise AssertionError(
                        f"Lower bound ({self.lower_bound}) must be less than or equal"
                        " to upper bound ({self.upper_bound})."
                    )
                if self.lower_bound == self.upper_bound and (
                    self.lower_bound_censor != ComparisonOperator.GTE
                    or self.upper_bound_censor != ComparisonOperator.STE
                ):
                    raise AssertionError(
                        f"Lower bound censor ({self.lower_bound_censor}) must be >="
                        " and upper bound censor ({self.upper_bound_censor}) must be"
                        " <= in case both bounds are equal."
                    )
        if self.lower_bound_censor is not None and self.lower_bound_censor not in {
            ComparisonOperator.GT,
            ComparisonOperator.GTE,
        }:
            raise AssertionError("Lower bound censor must be > or >=.")
        if self.upper_bound_censor is not None and self.upper_bound_censor not in {
            ComparisonOperator.ST,
            ComparisonOperator.STE,
        }:
            raise AssertionError("Upper bound censor must be < or <=.")

    @model_validator(mode="after")
    def _validate_state(self) -> Self:
        self._validate_state_bounds()
        # Generate the function to check if a value is within the range
        # The function is generated instead of defined to be able to optimize the check
        if self.lower_bound is not None and self.upper_bound is not None:
            if self.lower_bound == self.upper_bound:
                self._match = lambda x: x == self.lower_bound  # type: ignore
            elif (
                self.lower_bound_censor == ComparisonOperator.GTE
                and self.upper_bound_censor == ComparisonOperator.ST
            ):
                self._match = (  # type: ignore
                    lambda x: self.lower_bound <= x < self.upper_bound  # type: ignore
                )
            elif (
                self.lower_bound_censor == ComparisonOperator.GTE
                and self.upper_bound_censor == ComparisonOperator.STE
            ):
                self._match = (  # type: ignore
                    lambda x: self.lower_bound <= x <= self.upper_bound  # type: ignore
                )
            elif (
                self.lower_bound_censor == ComparisonOperator.GT
                and self.upper_bound_censor == ComparisonOperator.ST
            ):
                self._match = (  # type: ignore
                    lambda x: self.lower_bound < x < self.upper_bound  # type: ignore
                )
            elif (
                self.lower_bound_censor == ComparisonOperator.GT
                and self.upper_bound_censor == ComparisonOperator.STE
            ):
                self._match = (  # type: ignore
                    lambda x: self.lower_bound < x <= self.upper_bound  # type: ignore
                )
        elif self.lower_bound is not None:
            if self.lower_bound_censor == ComparisonOperator.GTE:
                self._match = lambda x: self.lower_bound <= x  # type: ignore
            elif self.lower_bound_censor == ComparisonOperator.GT:
                self._match = lambda x: self.lower_bound < x  # type: ignore
        elif self.upper_bound is not None:
            if self.upper_bound_censor == ComparisonOperator.ST:
                self._match = lambda x: x < self.upper_bound  # type: ignore
            elif self.upper_bound_censor == ComparisonOperator.STE:
                self._match = lambda x: x <= self.upper_bound  # type: ignore
        return self

    def _match(self, value: Any) -> bool:
        # Function is implemented dynamically in _validate_state
        raise NotImplementedError()


class TypedRangeFilter(RangeFilter):
    type: Literal[FilterType.RANGE.value]



from typing import ClassVar
from uuid import UUID

from pydantic import Field

from gen_epix.casedb.domain.model.case.persistable import CaseTypeColSet, CaseTypeSet
from gen_epix.commondb.domain import model as common_model
from gen_epix.fastapp.domain import Entity, create_keys, create_links


class BaseCasePolicy(common_model.Model):
    data_collection_id: UUID = Field(
        description="The ID of the data collection. FOREIGN KEY"
    )
    data_collection: common_model.DataCollection | None = Field(
        default=None, description="The data collection"
    )
    case_type_set_id: UUID = Field(
        description="The ID of the case type set. FOREIGN KEY",
    )
    case_type_set: CaseTypeSet | None = Field(
        default=None, description="The case type set"
    )
    is_active: bool = Field(description="Whether the right is active")
    add_case: bool = Field(
        description="Whether cases may be added to the data collection"
    )
    remove_case: bool = Field(
        description="Whether cases may be removed from the data collection"
    )
    add_case_set: bool = Field(
        description="Whether case sets may be added to the data collection"
    )
    remove_case_set: bool = Field(
        description="Whether case sets may be removed from the data collection"
    )


class OrganizationAccessCasePolicy(BaseCasePolicy):
    """
    Stores the access rights of an organization to a particular data collection.
    If an organization does not have a policy to a data collection, it has no
    access rights to that data collection.

    The access rights are limited to the case types in the case type set. If a
    case type is not in the case type set, the organization has no access
    rights to that data collection for that case type.
    """

    ENTITY: ClassVar = Entity(
        snake_case_plural_name="organization_access_case_policies",
        table_name="organization_access_case_policy",
        persistable=True,
        keys=create_keys(
            {
                1: (
                    "organization_id",
                    "data_collection_id",
                )
            }
        ),
        links=create_links(
            {
                1: ("organization_id", common_model.Organization, "organization"),
                2: (
                    "data_collection_id",
                    common_model.DataCollection,
                    "data_collection",
                ),
                3: ("case_type_set_id", CaseTypeSet, "case_type_set"),
                4: (
                    "read_case_type_col_set_id",
                    CaseTypeColSet,
                    "read_case_type_col_set",
                ),
                5: (
                    "write_case_type_col_set_id",
                    CaseTypeColSet,
                    "write_case_type_col_set",
                ),
            }
        ),
    )
    organization_id: UUID = Field(description="The ID of the organization. FOREIGN KEY")
    organization: common_model.Organization | None = Field(
        default=None, description="The organization"
    )
    is_private: bool = Field(
        description="Whether the data collection is private, limited to the case types in the case type set. When true, add/remove case and add/remove case set are considered (i) as the right to create/delete a case or case set in this data collection (setting case.created_in_data_collection to this data collection) and (ii) as the right to share the case or case set further in other data collections. Deleting a case or case set is only allowed when it can or has been removed from all other data collections as well."
    )
    read_case_type_col_set_id: UUID | None = Field(
        default=None,
        description="The ID of the case type column set for which values can be read, limited to the case types in the case type set. If empty, there are no read rights. FOREIGN KEY",
    )
    read_case_type_col_set: CaseTypeColSet | None = Field(
        default=None, description="The case type column set with read access"
    )
    write_case_type_col_set_id: UUID | None = Field(
        default=None,
        description="The ID of the case type column set for which values can be updated, limited to the case types in the case type set.  If empty, there are no write rights. FOREIGN KEY",
    )
    write_case_type_col_set: CaseTypeColSet | None = Field(
        default=None, description="The case type column set with write access"
    )
    read_case_set: bool = Field(
        description="Whether case set be read, limited to the case types in the case type set"
    )
    write_case_set: bool = Field(
        description="Whether case set be updated, limited to the case types in the case type set"
    )


class UserAccessCasePolicy(BaseCasePolicy):
    """
    Stores the maximum access rights of a user to a particular data collection,
    analogous to the organization access case policy.

    The actual access rights of a user are derived as the intersection of their
    maximum access rights stored here, and the access rights of the organization
    to which they belong.
    """

    ENTITY: ClassVar = Entity(
        snake_case_plural_name="user_access_case_policies",
        table_name="user_access_case_policy",
        persistable=True,
        keys=create_keys(
            {
                1: (
                    "user_id",
                    "data_collection_id",
                )
            }
        ),
        links=create_links(
            {
                1: ("user_id", common_model.User, "user"),
                2: (
                    "data_collection_id",
                    common_model.DataCollection,
                    "data_collection",
                ),
                3: ("case_type_set_id", CaseTypeSet, "case_type_set"),
                4: (
                    "read_case_type_col_set_id",
                    CaseTypeColSet,
                    "read_case_type_col_set",
                ),
                5: (
                    "write_case_type_col_set_id",
                    CaseTypeColSet,
                    "write_case_type_col_set",
                ),
            }
        ),
    )
    user_id: UUID = Field(description="The ID of the user. FOREIGN KEY")
    user: common_model.User | None = Field(default=None, description="The user")
    read_case_type_col_set_id: UUID | None = Field(
        default=None,
        description="The ID of the case type column set for which values can be read, limited to the case types in the case type set.  If empty, there are no read rights. FOREIGN KEY",
    )
    read_case_type_col_set: CaseTypeColSet | None = Field(
        default=None, description="The case type column set with read access"
    )
    write_case_type_col_set_id: UUID | None = Field(
        default=None,
        description="The ID of the case type column set for which values can be updated, limited to the case types in the case type set.  If empty, there are no write rights. FOREIGN KEY",
    )
    write_case_type_col_set: CaseTypeColSet | None = Field(
        default=None, description="The case type column set with write access"
    )
    read_case_set: bool = Field(
        description="Whether case set be read, limited to the case types in the case type set"
    )
    write_case_set: bool = Field(
        description="Whether case set be updated, limited to the case types in the case type set"
    )


class OrganizationShareCasePolicy(BaseCasePolicy):
    """
    Stores any additional case or case set share rights of an organization to a
    particular data collection, if the case or case set is already in a
    particular other data collection.

    The share rights are limited to the case types in the case type set. If a
    case type is not in the case type set, the organization has no share rights
    to that data collection for that case type.
    """

    ENTITY: ClassVar = Entity(
        snake_case_plural_name="organization_share_case_policies",
        table_name="organization_share_case_policy",
        persistable=True,
        keys=create_keys(
            {
                1: (
                    "organization_id",
                    "data_collection_id",
                    "from_data_collection_id",
                )
            }
        ),
        links=create_links(
            {
                1: ("organization_id", common_model.Organization, "organization"),
                2: (
                    "data_collection_id",
                    common_model.DataCollection,
                    "data_collection",
                ),
                3: ("case_type_set_id", CaseTypeSet, "case_type_set"),
                4: (
                    "from_data_collection_id",
                    common_model.DataCollection,
                    "from_data_collection",
                ),
            }
        ),
    )
    organization_id: UUID = Field(description="The ID of the organization. FOREIGN KEY")
    organization: common_model.Organization | None = Field(
        default=None, description="The organization"
    )
    from_data_collection_id: UUID = Field(
        description="The ID of the data collection from which the case type set is shared. FOREIGN KEY"
    )
    from_data_collection: common_model.DataCollection | None = Field(
        default=None,
        description="The data collection from which the case type set is shared",
    )


class UserShareCasePolicy(BaseCasePolicy):
    """
    Stores the maximum share rights of a user to a particular data collection,
    analogous to the organization share case policy.

    The actual share rights of a user are derived as the intersection of their
    maximum share rights stored here, and the share rights of the organization
    to which they belong.
    """

    ENTITY: ClassVar = Entity(
        snake_case_plural_name="user_share_case_policies",
        table_name="user_share_case_policy",
        persistable=True,
        keys=create_keys(
            {
                1: (
                    "user_id",
                    "data_collection_id",
                    "from_data_collection_id",
                )
            }
        ),
        links=create_links(
            {
                1: ("user_id", common_model.User, "user"),
                2: (
                    "data_collection_id",
                    common_model.DataCollection,
                    "data_collection",
                ),
                3: ("case_type_set_id", CaseTypeSet, "case_type_set"),
                4: (
                    "from_data_collection_id",
                    common_model.DataCollection,
                    "from_data_collection",
                ),
            }
        ),
    )
    user_id: UUID = Field(description="The ID of the user. FOREIGN KEY")
    user: common_model.User | None = Field(default=None, description="The user")
    from_data_collection_id: UUID = Field(
        description="The ID of the data collection from which the case type set is shared. FOREIGN KEY"
    )
    from_data_collection: common_model.DataCollection | None = Field(
        default=None,
        description="The data collection from which the case type set is shared",
    )



import datetime

import sqlalchemy.orm as orm
from sqlalchemy.orm import Mapped, declarative_mixin

from gen_epix.commondb.domain import DOMAIN, enum, model
from gen_epix.commondb.repositories.sa_model.base import RowMetadataMixin
from gen_epix.commondb.repositories.sa_model.util import (
    create_mapped_column,
    create_table_args,
)

Base: type = orm.declarative_base(name=enum.ServiceType.SYSTEM.value)


@declarative_mixin
class OutageMixin(RowMetadataMixin):
    description: Mapped[str | None] = create_mapped_column(
        DOMAIN, model.Outage, "description"
    )
    active_from: Mapped[datetime.datetime | None] = create_mapped_column(
        DOMAIN, model.Outage, "active_from"
    )
    active_to: Mapped[datetime.datetime | None] = create_mapped_column(
        DOMAIN, model.Outage, "active_to"
    )
    visible_from: Mapped[datetime.datetime | None] = create_mapped_column(
        DOMAIN, model.Outage, "visible_from"
    )
    visible_to: Mapped[datetime.datetime | None] = create_mapped_column(
        DOMAIN, model.Outage, "visible_to"
    )
    is_active: Mapped[bool | None] = create_mapped_column(
        DOMAIN, model.Outage, "is_active"
    )
    is_visible: Mapped[bool | None] = create_mapped_column(
        DOMAIN, model.Outage, "is_visible"
    )


class Outage(Base, OutageMixin):
    __tablename__, __table_args__ = create_table_args(model.Outage)



from fastapi import Request
from slowapi import Limiter
from slowapi.util import get_ipaddr


def limiter_key_func(request: Request) -> str:
    """
    Key function for the rate limiter.
    """
    if "authorization" in request.headers:
        return request.headers["authorization"]
    return get_ipaddr(request)


limiter = Limiter(key_func=limiter_key_func, default_limits=["10/second"])
